# ASN.1 "universal" data types
import operator, sys, math
from pyasn1.type import base, tag, constraint, namedtype, namedval, tagmap
from pyasn1.codec.ber import eoo
from pyasn1.compat import octets
from pyasn1 import error

# "Simple" ASN.1 types (yet incomplete)

class Integer(base.AbstractSimpleAsn1Item):
    tagSet = baseTagSet = tag.initTagSet(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 0x02)
        )
    namedValues = namedval.NamedValues()
    def __init__(self, value=None, tagSet=None, subtypeSpec=None,
                 namedValues=None):
        if namedValues is None:
            self.__namedValues = self.namedValues
        else:
            self.__namedValues = namedValues
        base.AbstractSimpleAsn1Item.__init__(
            self, value, tagSet, subtypeSpec
            )

    def __repr__(self):
        if self.__namedValues is not self.namedValues:
            return '%s, %r)' % (base.AbstractSimpleAsn1Item.__repr__(self)[:-1], self.__namedValues)
        else:
            return base.AbstractSimpleAsn1Item.__repr__(self)

    def __and__(self, value): return self.clone(self._value & value)
    def __rand__(self, value): return self.clone(value & self._value)
    def __or__(self, value): return self.clone(self._value | value)
    def __ror__(self, value): return self.clone(value | self._value)
    def __xor__(self, value): return self.clone(self._value ^ value)
    def __rxor__(self, value): return self.clone(value ^ self._value)
    def __lshift__(self, value): return self.clone(self._value << value)
    def __rshift__(self, value): return self.clone(self._value >> value)

    def __add__(self, value): return self.clone(self._value + value)
    def __radd__(self, value): return self.clone(value + self._value)
    def __sub__(self, value): return self.clone(self._value - value)
    def __rsub__(self, value): return self.clone(value - self._value)
    def __mul__(self, value): return self.clone(self._value * value)
    def __rmul__(self, value): return self.clone(value * self._value)
    def __mod__(self, value): return self.clone(self._value % value)
    def __rmod__(self, value): return self.clone(value % self._value)
    def __pow__(self, value, modulo=None): return self.clone(pow(self._value, value, modulo))
    def __rpow__(self, value): return self.clone(pow(value, self._value))

    if sys.version_info[0] <= 2:
        def __div__(self, value):  return self.clone(self._value // value)
        def __rdiv__(self, value):  return self.clone(value // self._value)
    else:
        def __truediv__(self, value):  return self.clone(self._value / value)
        def __rtruediv__(self, value):  return self.clone(value / self._value)
        def __divmod__(self, value):  return self.clone(self._value // value)
        def __rdivmod__(self, value):  return self.clone(value // self._value)

        __hash__ = base.AbstractSimpleAsn1Item.__hash__

    def __int__(self): return int(self._value)
    if sys.version_info[0] <= 2:
        def __long__(self): return long(self._value)
    def __float__(self): return float(self._value)    
    def __abs__(self): return self.clone(abs(self._value))
    def __index__(self): return int(self._value)
    def __pos__(self): return self.clone(+self._value)
    def __neg__(self): return self.clone(-self._value)
    def __invert__(self): return self.clone(~self._value)
    def __round__(self, n=0):
        r = round(self._value, n)
        if n:
            return self.clone(r)
        else:
            return r
    def __floor__(self): return math.floor(self._value)
    def __ceil__(self): return math.ceil(self._value)
    if sys.version_info[0:2] > (2, 5):
        def __trunc__(self): return self.clone(math.trunc(self._value))

    def __lt__(self, value): return self._value < value
    def __le__(self, value): return self._value <= value
    def __eq__(self, value): return self._value == value
    def __ne__(self, value): return self._value != value
    def __gt__(self, value): return self._value > value
    def __ge__(self, value): return self._value >= value

    def prettyIn(self, value):
        if not isinstance(value, str):
            try:
                return int(value)
            except:
                raise error.PyAsn1Error(
                    'Can\'t coerce %r into integer: %s' % (value, sys.exc_info()[1])
                    )
        r = self.__namedValues.getValue(value)
        if r is not None:
            return r
        try:
            return int(value)
        except:
            raise error.PyAsn1Error(
                'Can\'t coerce %r into integer: %s' % (value, sys.exc_info()[1])
                )

    def prettyOut(self, value):
        r = self.__namedValues.getName(value)
        return r is None and str(value) or repr(r)

    def getNamedValues(self): return self.__namedValues

    def clone(self, value=None, tagSet=None, subtypeSpec=None,
              namedValues=None):
        if value is None and tagSet is None and subtypeSpec is None \
               and namedValues is None:
            return self
        if value is None:
            value = self._value
        if tagSet is None:
            tagSet = self._tagSet
        if subtypeSpec is None:
            subtypeSpec = self._subtypeSpec
        if namedValues is None:
            namedValues = self.__namedValues
        return self.__class__(value, tagSet, subtypeSpec, namedValues)

    def subtype(self, value=None, implicitTag=None, explicitTag=None,
                subtypeSpec=None, namedValues=None):
        if value is None:
            value = self._value
        if implicitTag is not None:
            tagSet = self._tagSet.tagImplicitly(implicitTag)
        elif explicitTag is not None:
            tagSet = self._tagSet.tagExplicitly(explicitTag)
        else:
            tagSet = self._tagSet
        if subtypeSpec is None:
            subtypeSpec = self._subtypeSpec
        else:
            subtypeSpec = subtypeSpec + self._subtypeSpec
        if namedValues is None:
            namedValues = self.__namedValues
        else:
            namedValues = namedValues + self.__namedValues
        return self.__class__(value, tagSet, subtypeSpec, namedValues)

class Boolean(Integer):
    tagSet = baseTagSet = tag.initTagSet(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 0x01),
        )
    subtypeSpec = Integer.subtypeSpec+constraint.SingleValueConstraint(0,1)
    namedValues = Integer.namedValues.clone(('False', 0), ('True', 1))

class BitString(base.AbstractSimpleAsn1Item):
    tagSet = baseTagSet = tag.initTagSet(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 0x03)
        )
    namedValues = namedval.NamedValues()
    def __init__(self, value=None, tagSet=None, subtypeSpec=None,
                 namedValues=None):
        if namedValues is None:
            self.__namedValues = self.namedValues
        else:
            self.__namedValues = namedValues
        base.AbstractSimpleAsn1Item.__init__(
            self, value, tagSet, subtypeSpec
            )

    def clone(self, value=None, tagSet=None, subtypeSpec=None,
              namedValues=None):
        if value is None and tagSet is None and subtypeSpec is None \
               and namedValues is None:
            return self
        if value is None:
            value = self._value
        if tagSet is None:
            tagSet = self._tagSet
        if subtypeSpec is None:
            subtypeSpec = self._subtypeSpec
        if namedValues is None:
            namedValues = self.__namedValues
        return self.__class__(value, tagSet, subtypeSpec, namedValues)

    def subtype(self, value=None, implicitTag=None, explicitTag=None,
                subtypeSpec=None, namedValues=None):
        if value is None:
            value = self._value
        if implicitTag is not None:
            tagSet = self._tagSet.tagImplicitly(implicitTag)
        elif explicitTag is not None:
            tagSet = self._tagSet.tagExplicitly(explicitTag)
        else:
            tagSet = self._tagSet
        if subtypeSpec is None:
            subtypeSpec = self._subtypeSpec
        else:
            subtypeSpec = subtypeSpec + self._subtypeSpec
        if namedValues is None:
            namedValues = self.__namedValues
        else:
            namedValues = namedValues + self.__namedValues
        return self.__class__(value, tagSet, subtypeSpec, namedValues)

    def __str__(self): return str(tuple(self))

    # Immutable sequence object protocol

    def __len__(self):
        if self._len is None:
            self._len = len(self._value)
        return self._len
    def __getitem__(self, i):
        if isinstance(i, slice):
            return self.clone(operator.getitem(self._value, i))
        else:
            return self._value[i]

    def __add__(self, value): return self.clone(self._value + value)
    def __radd__(self, value): return self.clone(value + self._value)
    def __mul__(self, value): return self.clone(self._value * value)
    def __rmul__(self, value): return self * value

    def prettyIn(self, value):
        r = []
        if not value:
            return ()
        elif isinstance(value, str):
            if value[0] == '\'':
                if value[-2:] == '\'B':
                    for v in value[1:-2]:
                        if v == '0':
                            r.append(0)
                        elif v == '1':
                            r.append(1)
                        else:
                            raise error.PyAsn1Error(
                                'Non-binary BIT STRING initializer %s' % (v,)
                                )
                    return tuple(r)
                elif value[-2:] == '\'H':
                    for v in value[1:-2]:
                        i = 4
                        v = int(v, 16)
                        while i:
                            i = i - 1
                            r.append((v>>i)&0x01)
                    return tuple(r)
                else:
                    raise error.PyAsn1Error(
                        'Bad BIT STRING value notation %s' % (value,)
                        )                
            else:
                for i in value.split(','):
                    j = self.__namedValues.getValue(i)
                    if j is None:
                        raise error.PyAsn1Error(
                            'Unknown bit identifier \'%s\'' % (i,)
                            )
                    if j >= len(r):
                        r.extend([0]*(j-len(r)+1))
                    r[j] = 1
                return tuple(r)
        elif isinstance(value, (tuple, list)):
            r = tuple(value)
            for b in r:
                if b and b != 1:
                    raise error.PyAsn1Error(
                        'Non-binary BitString initializer \'%s\'' % (r,)
                        )
            return r
        elif isinstance(value, BitString):
            return tuple(value)
        else:
            raise error.PyAsn1Error(
                'Bad BitString initializer type \'%s\'' % (value,)
                )

    def prettyOut(self, value):
        return '\"\'%s\'B\"' % ''.join([str(x) for x in value])

try:
    all
except NameError:  # Python 2.4
    def all(iterable):
        for element in iterable:
            if not element:
                return False
        return True

class OctetString(base.AbstractSimpleAsn1Item):
    tagSet = baseTagSet = tag.initTagSet(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 0x04)
        )
    defaultBinValue = defaultHexValue = base.noValue
    encoding = 'us-ascii'
    def __init__(self, value=None, tagSet=None, subtypeSpec=None,
                 encoding=None, binValue=None, hexValue=None):
        if encoding is None:
            self._encoding = self.encoding
        else:
            self._encoding = encoding
        if binValue is not None:
            value = self.fromBinaryString(binValue)
        if hexValue is not None:
            value = self.fromHexString(hexValue)
        if value is None or value is base.noValue:
            value = self.defaultHexValue
        if value is None or value is base.noValue:
            value = self.defaultBinValue
        self.__asNumbersCache = None
        base.AbstractSimpleAsn1Item.__init__(self, value, tagSet, subtypeSpec)

    def clone(self, value=None, tagSet=None, subtypeSpec=None,
              encoding=None, binValue=None, hexValue=None):
        if value is None and tagSet is None and subtypeSpec is None and \
               encoding is None and binValue is None and hexValue is None:
            return self
        if value is None and binValue is None and hexValue is None:
            value = self._value
        if tagSet is None:
            tagSet = self._tagSet
        if subtypeSpec is None:
            subtypeSpec = self._subtypeSpec
        if encoding is None:
            encoding = self._encoding
        return self.__class__(
            value, tagSet, subtypeSpec, encoding, binValue, hexValue
            )
   
    if sys.version_info[0] <= 2:
        def prettyIn(self, value):
            if isinstance(value, str):
                return value
            elif isinstance(value, unicode):
                try:
                    return value.encode(self._encoding)
                except (LookupError, UnicodeEncodeError):
                    raise error.PyAsn1Error(
                        'Can\'t encode string \'%s\' with \'%s\' codec' % (value, self._encoding)
                    )
            elif isinstance(value, (tuple, list)):
                try:
                    return ''.join([ chr(x) for x in value ])
                except ValueError:
                    raise error.PyAsn1Error(
                        'Bad OctetString initializer \'%s\'' % (value,)
                    )                
            else:
                return str(value)
    else:
        def prettyIn(self, value):
            if isinstance(value, bytes):
                return value
            elif isinstance(value, str):
                try:
                    return value.encode(self._encoding)
                except UnicodeEncodeError:
                    raise error.PyAsn1Error(
                        'Can\'t encode string \'%s\' with \'%s\' codec' % (value, self._encoding)
                    )
            elif isinstance(value, OctetString):
                return value.asOctets()
            elif isinstance(value, (tuple, list, map)):
                try:
                    return bytes(value)
                except ValueError:
                    raise error.PyAsn1Error(
                        'Bad OctetString initializer \'%s\'' % (value,)
                    )
            else:
                try:
                    return str(value).encode(self._encoding)
                except UnicodeEncodeError:
                    raise error.PyAsn1Error(
                        'Can\'t encode string \'%s\' with \'%s\' codec' % (value, self._encoding)
                    )
                        

    def fromBinaryString(self, value):
        bitNo = 8; byte = 0; r = ()
        for v in value:
            if bitNo:
                bitNo = bitNo - 1
            else:
                bitNo = 7
                r = r + (byte,)
                byte = 0
            if v == '0':
                v = 0
            elif v == '1':
                v = 1
            else:
                raise error.PyAsn1Error(
                    'Non-binary OCTET STRING initializer %s' % (v,)
                    )
            byte = byte | (v << bitNo)
        return octets.ints2octs(r + (byte,))
        
    def fromHexString(self, value):            
        r = p = ()
        for v in value:
            if p:
                r = r + (int(p+v, 16),)
                p = ()
            else:
                p = v
        if p:
            r = r + (int(p+'0', 16),)
        return octets.ints2octs(r)

    def prettyOut(self, value):
        if sys.version_info[0] <= 2:
            numbers = tuple(( ord(x) for x in value ))
        else:
            numbers = tuple(value)
        if all(x >= 32 and x <= 126 for x in numbers):
            return str(value)
        else:
            return '0x' + ''.join(( '%.2x' % x for x in numbers ))

    def __repr__(self):
        r = []
        doHex = False
        if self._value is not self.defaultValue:
            for x in self.asNumbers():
                if x < 32 or x > 126:
                    doHex = True
                    break
            if not doHex:
                r.append('%r' % (self._value,))
        if self._tagSet is not self.tagSet:
            r.append('tagSet=%r' % (self._tagSet,))
        if self._subtypeSpec is not self.subtypeSpec:
            r.append('subtypeSpec=%r' % (self._subtypeSpec,))
        if self.encoding is not self._encoding:
            r.append('encoding=%r' % (self._encoding,))
        if doHex:
            r.append('hexValue=%r' % ''.join([ '%.2x' % x for x in self.asNumbers() ]))
        return '%s(%s)' % (self.__class__.__name__, ', '.join(r))
                                
    if sys.version_info[0] <= 2:
        def __str__(self): return str(self._value)
        def __unicode__(self):
            return self._value.decode(self._encoding, 'ignore')
        def asOctets(self): return self._value
        def asNumbers(self):
            if self.__asNumbersCache is None:
                self.__asNumbersCache = tuple([ ord(x) for x in self._value ])
            return self.__asNumbersCache
    else:
        def __str__(self): return self._value.decode(self._encoding, 'ignore')
        def __bytes__(self): return self._value
        def asOctets(self): return self._value
        def asNumbers(self):
            if self.__asNumbersCache is None:
                self.__asNumbersCache = tuple(self._value)
            return self.__asNumbersCache
 
    # Immutable sequence object protocol
    
    def __len__(self):
        if self._len is None:
            self._len = len(self._value)
        return self._len
    def __getitem__(self, i):
        if isinstance(i, slice):
            return self.clone(operator.getitem(self._value, i))
        else:
            return self._value[i]

    def __add__(self, value): return self.clone(self._value + self.prettyIn(value))
    def __radd__(self, value): return self.clone(self.prettyIn(value) + self._value)
    def __mul__(self, value): return self.clone(self._value * value)
    def __rmul__(self, value): return self * value
    def __int__(self): return int(self._value)
    def __float__(self): return float(self._value)
    
class Null(OctetString):
    defaultValue = ''.encode()  # This is tightly constrained
    tagSet = baseTagSet = tag.initTagSet(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 0x05)
        )
    subtypeSpec = OctetString.subtypeSpec+constraint.SingleValueConstraint(''.encode())
    
if sys.version_info[0] <= 2:
    intTypes = (int, long)
else:
    intTypes = (int,)

numericTypes = intTypes + (float,)

class ObjectIdentifier(base.AbstractSimpleAsn1Item):
    tagSet = baseTagSet = tag.initTagSet(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 0x06)
        )
    def __add__(self, other): return self.clone(self._value + other)
    def __radd__(self, other): return self.clone(other + self._value)

    def asTuple(self): return self._value
    
    # Sequence object protocol
    
    def __len__(self):
        if self._len is None:
            self._len = len(self._value)
        return self._len
    def __getitem__(self, i):
        if isinstance(i, slice):
            return self.clone(
                operator.getitem(self._value, i)
                )
        else:
            return self._value[i]

    def __str__(self): return self.prettyPrint()
    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.prettyPrint())

    def index(self, suboid): return self._value.index(suboid)

    def isPrefixOf(self, value):
        """Returns true if argument OID resides deeper in the OID tree"""
        l = len(self)
        if l <= len(value):
            if self._value[:l] == value[:l]:
                return 1
        return 0

    def prettyIn(self, value):
        """Dotted -> tuple of numerics OID converter"""
        if isinstance(value, tuple):
            pass
        elif isinstance(value, ObjectIdentifier):
            return tuple(value)        
        elif octets.isStringType(value):
            r = []
            for element in [ x for x in value.split('.') if x != '' ]:
                try:
                    r.append(int(element, 0))
                except ValueError:
                    raise error.PyAsn1Error(
                        'Malformed Object ID %s at %s: %s' %
                        (str(value), self.__class__.__name__, sys.exc_info()[1])
                        )
            value = tuple(r)
        else:
            try:
                value = tuple(value)
            except TypeError:
                raise error.PyAsn1Error(
                        'Malformed Object ID %s at %s: %s' %
                        (str(value), self.__class__.__name__,sys.exc_info()[1])
                        )

        for x in value:
            if not isinstance(x, intTypes) or x < 0:
                raise error.PyAsn1Error(
                    'Invalid sub-ID in %s at %s' % (value, self.__class__.__name__)
                    )
    
        return value

    def prettyOut(self, value): return '.'.join([ str(x) for x in value ])
    
class Real(base.AbstractSimpleAsn1Item):
    binEncBase = None # binEncBase = 16 is recommended for large numbers
    try:
        _plusInf = float('inf')
        _minusInf = float('-inf')
        _inf = (_plusInf, _minusInf)
    except ValueError:
        # Infinity support is platform and Python dependent
        _plusInf = _minusInf = None
        _inf = ()

    tagSet = baseTagSet = tag.initTagSet(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 0x09)
        )

    def __normalizeBase10(self, value):
        m, b, e = value
        while m and m % 10 == 0:
            m = m / 10
            e = e + 1
        return m, b, e

    def prettyIn(self, value):
        if isinstance(value, tuple) and len(value) == 3:
            if not isinstance(value[0], numericTypes) or \
                    not isinstance(value[1], intTypes) or \
                    not isinstance(value[2], intTypes):
                raise error.PyAsn1Error('Lame Real value syntax: %s' % (value,))
            if isinstance(value[0], float) and \
                self._inf and value[0] in self._inf:
                return value[0]
            if value[1] not in (2, 10):
                raise error.PyAsn1Error(
                    'Prohibited base for Real value: %s' % (value[1],)
                    )
            if value[1] == 10:
                value = self.__normalizeBase10(value)
            return value
        elif isinstance(value, intTypes):
            return self.__normalizeBase10((value, 10, 0))
        elif isinstance(value, (str, float)):
            if isinstance(value, str):
                try:
                    value = float(value)
                except ValueError:
                    raise error.PyAsn1Error(
                        'Bad real value syntax: %s' % (value,)
                    )
            if self._inf and value in self._inf:
                return value
            else:
                e = 0
                while int(value) != value:
                    value = value * 10
                    e = e - 1
                return self.__normalizeBase10((int(value), 10, e))
        elif isinstance(value, Real):
            return tuple(value)
        raise error.PyAsn1Error(
            'Bad real value syntax: %s' % (value,)
            )
        
    def prettyOut(self, value):
        if value in self._inf:
            return '\'%s\'' % value
        else:
            return str(value)

    def prettyPrint(self, scope=0):
        if self.isInfinity():
            return self.prettyOut(self._value)
        else:
            return str(float(self))

    def isPlusInfinity(self): return self._value == self._plusInf
    def isMinusInfinity(self): return self._value == self._minusInf
    def isInfinity(self): return self._value in self._inf
    
    def __str__(self): return str(float(self))
    
    def __add__(self, value): return self.clone(float(self) + value)
    def __radd__(self, value): return self + value
    def __mul__(self, value): return self.clone(float(self) * value)
    def __rmul__(self, value): return self * value
    def __sub__(self, value): return self.clone(float(self) - value)
    def __rsub__(self, value): return self.clone(value - float(self))
    def __mod__(self, value): return self.clone(float(self) % value)
    def __rmod__(self, value): return self.clone(value % float(self))
    def __pow__(self, value, modulo=None): return self.clone(pow(float(self), value, modulo))
    def __rpow__(self, value): return self.clone(pow(value, float(self)))

    if sys.version_info[0] <= 2:
        def __div__(self, value): return self.clone(float(self) / value)
        def __rdiv__(self, value): return self.clone(value / float(self))
    else:
        def __truediv__(self, value): return self.clone(float(self) / value)
        def __rtruediv__(self, value): return self.clone(value / float(self))
        def __divmod__(self, value): return self.clone(float(self) // value)
        def __rdivmod__(self, value): return self.clone(value // float(self))

    def __int__(self): return int(float(self))
    if sys.version_info[0] <= 2:
        def __long__(self): return long(float(self))
    def __float__(self):
        if self._value in self._inf:
            return self._value
        else:
            return float(
                self._value[0] * pow(self._value[1], self._value[2])
            )
    def __abs__(self): return self.clone(abs(float(self)))
    def __pos__(self): return self.clone(+float(self))
    def __neg__(self): return self.clone(-float(self))
    def __round__(self, n=0):
        r = round(float(self), n)
        if n:
            return self.clone(r)
        else:
            return r
    def __floor__(self): return self.clone(math.floor(float(self)))
    def __ceil__(self): return self.clone(math.ceil(float(self)))
    if sys.version_info[0:2] > (2, 5):
        def __trunc__(self): return self.clone(math.trunc(float(self)))

    def __lt__(self, value): return float(self) < value
    def __le__(self, value): return float(self) <= value
    def __eq__(self, value): return float(self) == value
    def __ne__(self, value): return float(self) != value
    def __gt__(self, value): return float(self) > value
    def __ge__(self, value): return float(self) >= value

    if sys.version_info[0] <= 2:
        def __nonzero__(self): return bool(float(self))
    else:
        def __bool__(self): return bool(float(self))
        __hash__ = base.AbstractSimpleAsn1Item.__hash__

    def __getitem__(self, idx):
        if self._value in self._inf:
            raise error.PyAsn1Error('Invalid infinite value operation')
        else:
            return self._value[idx]
    
class Enumerated(Integer):
    tagSet = baseTagSet = tag.initTagSet(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 0x0A)
        )

# "Structured" ASN.1 types

class SetOf(base.AbstractConstructedAsn1Item):
    componentType = None
    tagSet = baseTagSet = tag.initTagSet(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatConstructed, 0x11)
        )
    typeId = 1
    strictConstraints = False

    def _cloneComponentValues(self, myClone, cloneValueFlag):
        idx = 0; l = len(self._componentValues)
        while idx < l:
            c = self._componentValues[idx]
            if c is not None:
                if isinstance(c, base.AbstractConstructedAsn1Item):
                    myClone.setComponentByPosition(
                        idx, c.clone(cloneValueFlag=cloneValueFlag)
                        )
                else:
                    myClone.setComponentByPosition(idx, c.clone())
            idx = idx + 1
        
    def _verifyComponent(self, idx, value):
        t = self._componentType
        if t is None:
            return
        if not t.isSameTypeWith(value,matchConstraints=self.strictConstraints):
            raise error.PyAsn1Error('Component value is tag-incompatible: %r vs %r' % (value, t))
        if self.strictConstraints and \
                not t.isSuperTypeOf(value, matchTags=False):
            raise error.PyAsn1Error('Component value is constraints-incompatible: %r vs %r' % (value, t))

    def getComponentByPosition(self, idx): return self._componentValues[idx]
    def setComponentByPosition(self, idx, value=None, verifyConstraints=True):
        l = len(self._componentValues)
        if idx >= l:
            self._componentValues = self._componentValues + (idx-l+1)*[None]
        if value is None:
            if self._componentValues[idx] is None:
                if self._componentType is None:
                    raise error.PyAsn1Error('Component type not defined')
                self._componentValues[idx] = self._componentType.clone()
                self._componentValuesSet = self._componentValuesSet + 1
            return self
        elif not isinstance(value, base.Asn1Item):
            if self._componentType is None:
                raise error.PyAsn1Error('Component type not defined')
            if isinstance(self._componentType, base.AbstractSimpleAsn1Item):
                value = self._componentType.clone(value=value)
            else:
                raise error.PyAsn1Error('Instance value required')
        if verifyConstraints:
            if self._componentType is not None:
                self._verifyComponent(idx, value)
            self._verifySubtypeSpec(value, idx)            
        if self._componentValues[idx] is None:
            self._componentValuesSet = self._componentValuesSet + 1
        self._componentValues[idx] = value
        return self

    def getComponentTagMap(self):
        if self._componentType is not None:
            return self._componentType.getTagMap()

    def prettyPrint(self, scope=0):
        scope = scope + 1
        r = self.__class__.__name__ + ':\n'        
        for idx in range(len(self._componentValues)):
            r = r + ' '*scope
            if self._componentValues[idx] is None:
                r = r + '<empty>'
            else:
                r = r + self._componentValues[idx].prettyPrint(scope)
        return r

    def prettyPrintType(self, scope=0):
        scope = scope + 1
        r = '%s -> %s {\n' % (self.getTagSet(), self.__class__.__name__)
        if self._componentType is not None:
            r = r + ' '*scope
            r = r + self._componentType.prettyPrintType(scope)
        return r + '\n' + ' '*(scope-1) + '}'

class SequenceOf(SetOf):
    tagSet = baseTagSet = tag.initTagSet(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatConstructed, 0x10)
        )
    typeId = 2

class SequenceAndSetBase(base.AbstractConstructedAsn1Item):
    componentType = namedtype.NamedTypes()
    strictConstraints = False
    def __init__(self, componentType=None, tagSet=None,
                 subtypeSpec=None, sizeSpec=None):
        if componentType is None:
            componentType = self.componentType
        base.AbstractConstructedAsn1Item.__init__(
            self, componentType.clone(), tagSet, subtypeSpec, sizeSpec
        )
        self._componentTypeLen = len(self._componentType)

    def __getitem__(self, idx):
        if isinstance(idx, str):
            return self.getComponentByName(idx)
        else:
            return base.AbstractConstructedAsn1Item.__getitem__(self, idx)

    def __setitem__(self, idx, value):
        if isinstance(idx, str):
            self.setComponentByName(idx, value)
        else:
            base.AbstractConstructedAsn1Item.__setitem__(self, idx, value)
        
    def _cloneComponentValues(self, myClone, cloneValueFlag):
        idx = 0; l = len(self._componentValues)
        while idx < l:
            c = self._componentValues[idx]
            if c is not None:
                if isinstance(c, base.AbstractConstructedAsn1Item):
                    myClone.setComponentByPosition(
                        idx, c.clone(cloneValueFlag=cloneValueFlag)
                        )
                else:
                    myClone.setComponentByPosition(idx, c.clone())
            idx = idx + 1

    def _verifyComponent(self, idx, value):
        if idx >= self._componentTypeLen:
            raise error.PyAsn1Error(
                'Component type error out of range'
                )
        t = self._componentType[idx].getType()
        if not t.isSameTypeWith(value,matchConstraints=self.strictConstraints):
            raise error.PyAsn1Error('Component value is tag-incompatible: %r vs %r' % (value, t))
        if self.strictConstraints and \
                not t.isSuperTypeOf(value, matchTags=False):
            raise error.PyAsn1Error('Component value is constraints-incompatible: %r vs %r' % (value, t))

    def getComponentByName(self, name):
        return self.getComponentByPosition(
            self._componentType.getPositionByName(name)
            )
    def setComponentByName(self, name, value=None, verifyConstraints=True):
        return self.setComponentByPosition(
            self._componentType.getPositionByName(name),value,verifyConstraints
        )

    def getComponentByPosition(self, idx):
        try:
            return self._componentValues[idx]
        except IndexError:
            if idx < self._componentTypeLen:
                return
            raise
    def setComponentByPosition(self, idx, value=None,
                               verifyConstraints=True,
                               exactTypes=False,
                               matchTags=True,
                               matchConstraints=True):
        l = len(self._componentValues)
        if idx >= l:
            self._componentValues = self._componentValues + (idx-l+1)*[None]
        if value is None:
            if self._componentValues[idx] is None:
                self._componentValues[idx] = self._componentType.getTypeByPosition(idx).clone()
                self._componentValuesSet = self._componentValuesSet + 1
            return self
        elif not isinstance(value, base.Asn1Item):
            t = self._componentType.getTypeByPosition(idx)
            if isinstance(t, base.AbstractSimpleAsn1Item):
                value = t.clone(value=value)
            else:
                raise error.PyAsn1Error('Instance value required')
        if verifyConstraints:
            if self._componentTypeLen:
                self._verifyComponent(idx, value)
            self._verifySubtypeSpec(value, idx)            
        if self._componentValues[idx] is None:
            self._componentValuesSet = self._componentValuesSet + 1
        self._componentValues[idx] = value
        return self

    def getNameByPosition(self, idx):
        if self._componentTypeLen:
            return self._componentType.getNameByPosition(idx)

    def getDefaultComponentByPosition(self, idx):
        if self._componentTypeLen and self._componentType[idx].isDefaulted:
            return self._componentType[idx].getType()

    def getComponentType(self):
        if self._componentTypeLen:
            return self._componentType
    
    def setDefaultComponents(self):
        if self._componentTypeLen == self._componentValuesSet:
            return
        idx = self._componentTypeLen
        while idx:
            idx = idx - 1
            if self._componentType[idx].isDefaulted:
                if self.getComponentByPosition(idx) is None:
                    self.setComponentByPosition(idx)
            elif not self._componentType[idx].isOptional:
                if self.getComponentByPosition(idx) is None:
                    raise error.PyAsn1Error(
                        'Uninitialized component #%s at %r' % (idx, self)
                        )

    def prettyPrint(self, scope=0):
        scope = scope + 1
        r = self.__class__.__name__ + ':\n'
        for idx in range(len(self._componentValues)):
            if self._componentValues[idx] is not None:
                r = r + ' '*scope
                componentType = self.getComponentType()
                if componentType is None:
                    r = r + '<no-name>'
                else:
                    r = r + componentType.getNameByPosition(idx)
                r = '%s=%s\n' % (
                    r, self._componentValues[idx].prettyPrint(scope)
                    )
        return r

    def prettyPrintType(self, scope=0):
        scope = scope + 1
        r = '%s -> %s {\n' % (self.getTagSet(), self.__class__.__name__)
        for idx in range(len(self.componentType)):
            r = r + ' '*scope
            r = r + '"%s"' % self.componentType.getNameByPosition(idx)
            r = '%s = %s\n' % (
                r, self._componentType.getTypeByPosition(idx).prettyPrintType(scope)
            )
        return r + '\n' + ' '*(scope-1) + '}'

class Sequence(SequenceAndSetBase):
    tagSet = baseTagSet = tag.initTagSet(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatConstructed, 0x10)
        )
    typeId = 3

    def getComponentTagMapNearPosition(self, idx):
        if self._componentType:
            return self._componentType.getTagMapNearPosition(idx)
    
    def getComponentPositionNearType(self, tagSet, idx):
        if self._componentType:
            return self._componentType.getPositionNearType(tagSet, idx)
        else:
            return idx
    
class Set(SequenceAndSetBase):
    tagSet = baseTagSet = tag.initTagSet(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatConstructed, 0x11)
        )
    typeId = 4

    def getComponent(self, innerFlag=0): return self
    
    def getComponentByType(self, tagSet, innerFlag=0):
        c = self.getComponentByPosition(
            self._componentType.getPositionByType(tagSet)
            )
        if innerFlag and isinstance(c, Set):
            # get inner component by inner tagSet
            return c.getComponent(1)
        else:
            # get outer component by inner tagSet
            return c
        
    def setComponentByType(self, tagSet, value=None, innerFlag=0,
                           verifyConstraints=True):
        idx = self._componentType.getPositionByType(tagSet)
        t = self._componentType.getTypeByPosition(idx)
        if innerFlag:  # set inner component by inner tagSet
            if t.getTagSet():
                return self.setComponentByPosition(
                    idx, value, verifyConstraints
                )
            else:
                t = self.setComponentByPosition(idx).getComponentByPosition(idx)
                return t.setComponentByType(
                    tagSet, value, innerFlag, verifyConstraints
                )
        else:  # set outer component by inner tagSet
            return self.setComponentByPosition(
                idx, value, verifyConstraints
            )
            
    def getComponentTagMap(self):
        if self._componentType:
            return self._componentType.getTagMap(True)

    def getComponentPositionByType(self, tagSet):
        if self._componentType:
            return self._componentType.getPositionByType(tagSet)

class Choice(Set):
    tagSet = baseTagSet = tag.TagSet()  # untagged
    sizeSpec = constraint.ConstraintsIntersection(
        constraint.ValueSizeConstraint(1, 1)
        )
    typeId = 5
    _currentIdx = None

    def __eq__(self, other):
        if self._componentValues:
            return self._componentValues[self._currentIdx] == other
        return NotImplemented
    def __ne__(self, other):
        if self._componentValues:
            return self._componentValues[self._currentIdx] != other
        return NotImplemented
    def __lt__(self, other):
        if self._componentValues:
            return self._componentValues[self._currentIdx] < other
        return NotImplemented
    def __le__(self, other):
        if self._componentValues:
            return self._componentValues[self._currentIdx] <= other
        return NotImplemented
    def __gt__(self, other):
        if self._componentValues:
            return self._componentValues[self._currentIdx] > other
        return NotImplemented
    def __ge__(self, other):
        if self._componentValues:
            return self._componentValues[self._currentIdx] >= other
        return NotImplemented
    if sys.version_info[0] <= 2:
        def __nonzero__(self): return bool(self._componentValues)
    else:
        def __bool__(self): return bool(self._componentValues)

    def __len__(self): return self._currentIdx is not None and 1 or 0
    
    def verifySizeSpec(self):
        if self._currentIdx is None:
            raise error.PyAsn1Error('Component not chosen')
        else:
            self._sizeSpec(' ')

    def _cloneComponentValues(self, myClone, cloneValueFlag):
        try:
            c = self.getComponent()
        except error.PyAsn1Error:
            pass
        else:
            if isinstance(c, Choice):
                tagSet = c.getEffectiveTagSet()
            else:
                tagSet = c.getTagSet()
            if isinstance(c, base.AbstractConstructedAsn1Item):
                myClone.setComponentByType(
                    tagSet, c.clone(cloneValueFlag=cloneValueFlag)
                    )
            else:
                myClone.setComponentByType(tagSet, c.clone())

    def setComponentByPosition(self, idx, value=None, verifyConstraints=True):
        l = len(self._componentValues)
        if idx >= l:
            self._componentValues = self._componentValues + (idx-l+1)*[None]
        if self._currentIdx is not None:
            self._componentValues[self._currentIdx] = None
        if value is None:
            if self._componentValues[idx] is None:
                self._componentValues[idx] = self._componentType.getTypeByPosition(idx).clone()
                self._componentValuesSet = 1
                self._currentIdx = idx
            return self
        elif not isinstance(value, base.Asn1Item):
            value = self._componentType.getTypeByPosition(idx).clone(
                value=value
                )
        if verifyConstraints:
            if self._componentTypeLen:
                self._verifyComponent(idx, value)
            self._verifySubtypeSpec(value, idx)            
        self._componentValues[idx] = value
        self._currentIdx = idx
        self._componentValuesSet = 1
        return self

    def getMinTagSet(self):
        if self._tagSet:
            return self._tagSet
        else:
            return self._componentType.genMinTagSet()

    def getEffectiveTagSet(self):
        if self._tagSet:
            return self._tagSet
        else:
            c = self.getComponent()
            if isinstance(c, Choice):
                return c.getEffectiveTagSet()
            else:
                return c.getTagSet()

    def getTagMap(self):
        if self._tagSet:
            return Set.getTagMap(self)
        else:
            return Set.getComponentTagMap(self)

    def getComponent(self, innerFlag=0):
        if self._currentIdx is None:
            raise error.PyAsn1Error('Component not chosen')
        else:
            c = self._componentValues[self._currentIdx]
            if innerFlag and isinstance(c, Choice):
                return c.getComponent(innerFlag)
            else:
                return c

    def getName(self, innerFlag=0):
        if self._currentIdx is None:
            raise error.PyAsn1Error('Component not chosen')
        else:
            if innerFlag:
                c = self._componentValues[self._currentIdx]
                if isinstance(c, Choice):
                    return c.getName(innerFlag)
            return self._componentType.getNameByPosition(self._currentIdx)

    def setDefaultComponents(self): pass

class Any(OctetString):
    tagSet = baseTagSet = tag.TagSet()  # untagged
    typeId = 6

    def getTagMap(self):
        return tagmap.TagMap(
            { self.getTagSet(): self },
            { eoo.endOfOctets.getTagSet(): eoo.endOfOctets },
            self
            )

# XXX
# coercion rules?



import json

from xml.etree.ElementTree import Element
from xml.etree import cElementTree as eTree

from sysactions import (
    StopAction,
    get_model,
    get_operator_session,
    show_messagebox,
)
from typing import (
    Any,
    List,
    Dict,
    Optional,
    Union,
)

from actions.order import update_tenders
from utilfunctions import get_device_service

NOT_ELECTRONIC_TYPE_ID = (0, 9)


def has_electronic_payment(
    order,                          # type: Optional[Element]
    tender_electronic_type=None,    # type: Optional[int]
):
    # type: (...) -> bool

    if tender_electronic_type is not None:
        return tender_electronic_type != 'NON_ELECTRONIC'

    if not order:
        return False

    tenders = order.findall("TenderHistory/Tender")
    for tender in tenders:
        if _get_electronic_type_id(tender) not in NOT_ELECTRONIC_TYPE_ID:
            return True

    return False


def get_electronic_tenders(
    order,  # type: Element
):
    # type: (...) -> List

    tenders = order.findall("TenderHistory/Tender")
    ret_tenders = []
    for tender in tenders:
        electronic_type = _get_electronic_type_id(tender)
        if electronic_type not in NOT_ELECTRONIC_TYPE_ID:
            ret_tenders.append({
                "id": int(tender.get("tenderId")),
                "type": int(tender.get("tenderType")),
                "amount": tender.get("tenderAmount"),
                "electronicType": electronic_type,
                "detail": json.loads(tender.get("tenderDetail", "{}"))
            })

    return ret_tenders


def _get_electronic_type_id(
    tender,     # type: Element
):
    # type: (...) -> int

    return int(tender.get("electronicType") or tender.get("eletronicType"))


def get_tender_details(
    payment_data,   # type: Dict
):
    # type: (...) -> str

    tender_details = json.dumps({
        "Media": payment_data.get("Media", "").strip(),
        "CNPJAuth": payment_data["CNPJAuth"],
        "TransactionProcessor": payment_data["TransactionProcessor"],
        "Bandeira": payment_data["Bandeira"],
        "IdAuth": payment_data["IdAuth"],
        "AuthCode": payment_data["AuthCode"],
        "ReceiptCustomer": payment_data["ReceiptCustomer"],
        "ReceiptMerchant": payment_data["ReceiptMerchant"],
        "AccreditorSat": payment_data["AccreditorSat"] if "AccreditorSat" in payment_data else None,
        "NfceBrand": payment_data["NfceBrand"] if "NfceBrand" in payment_data else None,
        "NsuHost": payment_data["NsuHost"] if "NsuHost" in payment_data else None,
        "CardNumber": payment_data["CardNumber"] if "CardNumber" in payment_data else None,
        "TransactionIdentifier": payment_data.get("TransactionIdentifier"),
        "OriginalNfceBrand": payment_data["OriginalNfceBrand"] if "OriginalNfceBrand" in payment_data else None
    })

    return tender_details


def check_eft_transaction(
    model,              # type: eTree.XML
    pos_id,             # type: str
    order_id,           # type: str
    eft_xml,            # type: str
    electronic_type,    # type: str
):
    # type: (...) -> None

    if eft_xml is None or eTree.XML(eft_xml).get("Result") != "0":
        if eft_xml is None:
            show_messagebox(pos_id, "$ERROR_PROCESSING_TEF")
            raise StopAction()

        if eft_xml:
            xml = eTree.XML(eft_xml)
            response = xml.get("Result") or "(none)"
            if electronic_type in ["CREDIT_CARD", "DEBIT_CARD"]:
                message = "${}_NOT_PROCESSED|{}".format(electronic_type, response.encode('utf-8'))
                show_messagebox(
                    pos_id,
                    message=message
                )

            raise StopAction()


def get_tender_id(
    all_tenders,                # type: eTree
    is_table_service=False,     # type: Optional[bool]
):
    # type: (...) -> str

    if not is_table_service:
        return ""

    if len(all_tenders) == 0:
        return "1"

    last_tender_id = all_tenders[-1].get("tenderId", 0)
    return str(int(last_tender_id) + 1)


def save_cancel_receipt(
    pos_id,     # type: str
    order,      # type: eTree.XML
    tender_id,  # type: str
    data,       # type: str
):
    # type: (...) -> bool

    new_tender_detail = _get_cancel_receipt(
        order,
        int(tender_id),
        json.loads(data)
    )

    success_update_tenders = update_tenders(
        pos_id,
        order.get("orderId"),
        tender_id,
        json.dumps(new_tender_detail)
    )

    return success_update_tenders


def _get_cancel_receipt(
    order,          # type: eTree
    tender_id,      # type: int
    cancel_data,    # type: Dict
):
    # type: (...) -> List

    for original_tender in order.findall('TenderHistory/Tender'):
        current_tender_id = int(original_tender.get('tenderId'))
        if current_tender_id == tender_id:
            tender = {}
            tender_detail = json.loads(original_tender.get("tenderDetail", {}))
            tender_detail["receiptCancel"] = cancel_data
            tender['tenderdetail'] = json.dumps(tender_detail)
            tender['tenderid'] = current_tender_id
            return [tender]


def get_electronic_payment_component_name(
    pos_id,     # type: Any[Union[str, int]]
):
    # type: (...) -> Optional[str]

    return get_device_service(
        pos_id=pos_id,
        device_type="eft"
    )




# encoding: utf-8

import json

from datetime import datetime
from typing import (
    Any,
    Dict,
    List,
    Tuple,
)

from old_helper import BaseRepository


class BenefitsOrderRepository(BaseRepository):

    def __init__(self, mb_context, pos_list):
        super(BenefitsOrderRepository, self).__init__(mb_context)

        self.pos_list = pos_list

    def get_benefits_orders_detailed(
            self,
            initial_date,   # type: str
            end_date,       # type: str
            operator_id,    # type: str
            benefits        # type: Dict
    ):
        # type: (...) -> List[Dict[str, Any]]

        query = self._build_query(initial_date, end_date, operator_id)

        def inner_func(conn):
            def get_product_code(
                actions,    # type: Dict
            ):
                # type: (...) -> str

                for item in actions:
                    if 'item' not in item and 'ByGroups' not in item and type(actions[item]) in [list, dict] :
                        return get_product_code(actions[item])
                    elif 'item' in item:
                        return item['item']
                    else:
                        return ""

            def get_product_benefit(
                product_code,   # type: str
                order_id,       # type: str
            ):
                # type: (...) -> (str, str)

                product_benefit_query = """
                    SELECT
                        op.DefaultUnitPrice,
                        p.ProductName,
                        (
                            SELECT
                                SUM(oi.OrderedQty)
                            FROM OrderItem oi
                            WHERE oi.OrderId='{1}' AND oi.PartCode='{0}'
                        ) as Qty
                    FROM productdb.Price OP
                    INNER JOIN productdb.Product p
                    WHERE
                        op.ProductCode = '{0}'
                        AND p.ProductCode = '{0}'
                    LIMIT 1;
                """
                if product_code:
                    for new_row in conn.select(product_benefit_query.format(product_code, order_id)):
                        product_name = new_row.get_entry('ProductName')
                        default_unit_price = new_row.get_entry('DefaultUnitPrice')
                        qty = new_row.get_entry('Qty')
                        return product_name, default_unit_price, qty
                return None, None, None

            benefits_orders = []
            chosen_benefits = [x.get("key") for x in benefits]
            for row in conn.select(query):
                benefit_list_col = json.loads(row.get_entry("BenefitList"))
                benefit_list = json.loads(benefit_list_col.get("benefitList", []))
                for benefit in benefit_list:
                    benefit_name = benefit.get("benefit").get("name")

                    if benefit.get("benefit_applied") and benefit_name in chosen_benefits:
                        benefit_product_code = get_product_code(actions=benefit['benefit']['actions'])
                        benefit_product_name, benefit_product_price, product_qty = get_product_benefit(
                            product_code=benefit_product_code,
                            order_id=row.get_entry("OrderId")
                        )
                        if not benefit_product_name:
                            benefit_product_name = row.get_entry("ProductName")
                            benefit_product_price = float(row.get_entry("ProductUnitPrice"))
                            product_qty = row.get_entry("OrderedQty")

                        session_id = row.get_entry("SessionId")
                        operator = int(session_id.split("user=")[1].split(",")[0]) if session_id else ""
                        discount_type = [self._label_discount_type(a) for a in benefit['benefit']['actions'].items()]
                        benefits_orders.append({
                            "business_period": row.get_entry("BusinessPeriod"),
                            "order_time": row.get_entry("OrderTime"),
                            "operator": operator,
                            "order_id": row.get_entry("OrderId"),
                            "ordered_qty": int(float(product_qty)),
                            "product_unit_price": float(benefit_product_price),
                            "product_name": benefit_product_name,
                            "benefit_total_value": benefit["benefit_total_value"],
                            "discount_type": discount_type[0] or "",
                            "total_gross": float(row.get_entry("TotalGross"))
                        })
            return benefits_orders

        return self.execute_parallel_in_all_databases_returning_flat_list(inner_func, self.pos_list)

    @staticmethod
    def _label_discount_type(
        action,     # type: Tuple
    ):
        # type: (...) -> str

        action_name = action[1].items()[0][0].lower()
        action_value = action[1].items()[0][1]
        action_value = action_value if isinstance(action_value, float) else float(action_value[0].get('percentage', 0))
        action_value = int(action_value) if int(action_value * 10 % 10) == 0 else action_value

        return "$VALUE" if 'amount' in action_name else "$PERCENTAGE ({}%)".format(action_value)

    def get_benefits_orders_consolidated(
            self,
            initial_date,   # type: str
            end_date,       # type: str
            operator_id     # type: str
    ):
        # type: (...) -> List[Dict[str, Dict]]

        query = self._build_query(initial_date, end_date, operator_id)

        def inner_func(conn):
            benefits_orders = []
            for row in conn.select(query):
                benefit_list_col = json.loads(row.get_entry("BenefitList"))
                benefit_list = json.loads(benefit_list_col.get("benefitList", []))

                data = {}
                for benefit in benefit_list:
                    if benefit.get("benefit_applied"):
                        benefit_id = benefit.get("benefit_id")
                        item_unit_price = benefit.get("discount_item_values", 0.0)
                        if item_unit_price:
                            item_unit_price = item_unit_price[str(benefit.get("used_discount_ids")[0])]
                            item_unit_price = item_unit_price[item_unit_price.keys()[0]]
                        data.setdefault(
                            benefit_id,
                            {
                                "quantity": 0,
                                "product_unit_price": item_unit_price,
                                "benefit_total_value": 0.0,
                                "benefit_user_description": (
                                    benefit.get("benefit", {}).get("userDescription")
                                    or benefit_id
                                )
                            }
                        )
                        data[benefit_id]["quantity"] += len(benefit.get("discount_item_values"))
                        data[benefit_id]["benefit_total_value"] += benefit.get("benefit_total_value", 0.0)

                benefits_orders.append(data)

            return benefits_orders

        return self.execute_parallel_in_all_databases_returning_flat_list(inner_func, self.pos_list)

    def _build_query(
            self,
            initial_date,  # type: str
            end_date,  # type: str
            operator_id  # type: str
    ):
        # type: (...) -> str

        initial_date = datetime.strptime(initial_date, "%Y%m%d").date()
        end_date = datetime.strptime(end_date, "%Y%m%d").date()

        conditional = self._QUERY_ORDERS_CONDITIONAL_WITHOUT_OPERATOR.format(
            initial_date=initial_date,
            end_date=end_date
        )

        if operator_id is not None:
            operator_id = int(operator_id)
            conditional += self._QUERY_ORDERS_CONDITIONAL_WITH_OPERATOR.format(
                initial_date=initial_date,
                end_date=end_date,
                operator_id=operator_id
            )

        parts = [
            self._QUERY_ORDERS_PREAMBLE,
            conditional,
            self._QUERY_ORDERS_EPILOGUE
        ]
        return "".join(parts)

    _QUERY_ORDERS_PREAMBLE = """
        SELECT
            o.OrderId,
            o.TotalGross,
            o.SessionId,
            o.BusinessPeriod,
            od.OrderTime,
            oi.OrderedQty,
            ocp.Value AS BenefitList,
            p.ProductName,
            COALESCE(
                oi.OverwrittenUnitPrice,
                op.DefaultUnitPrice, 
                0.0
            ) AS ProductUnitPrice
        FROM OrderItem oi
        INNER JOIN Orders o ON o.OrderId = oi.OrderId
        INNER JOIN OrderCustomProperties ocp ON ocp.OrderId = o.OrderId
        INNER JOIN OrderPrice op ON op.OrderId = oi.OrderId
        INNER JOIN productdb.Product p ON p.ProductCode = oi.PartCode
        INNER JOIN (SELECT OrderId, 
                           strftime('%H:%M:%S', max(Timestamp), 'localtime') OrderTime
				      FROM OrderStateHistory
					 WHERE StateId = 5
					 GROUP BY OrderId
					) od on o.OrderId = od.OrderId
    """

    _QUERY_ORDERS_CONDITIONAL_WITH_OPERATOR = """
            AND o.SessionId LIKE '%user={operator_id:d},%'
    """

    _QUERY_ORDERS_CONDITIONAL_WITHOUT_OPERATOR = """
        WHERE o.StateId = 5 -- PAID
            AND ocp.Key = 'BENEFIT_LIST'
            AND o.BusinessPeriod BETWEEN '{initial_date:%Y%m%d}' AND '{end_date:%Y%m%d}'
    """

    _QUERY_ORDERS_EPILOGUE = """
        GROUP BY oi.OrderId
        ORDER BY o.OrderId, o.BusinessPeriod, od.OrderTime, p.ProductName
    """



# -*- coding: utf-8 -*-
import os
import sys
import base64

from fiscalpersistence import Tender
from typing import Optional, Dict, List

import msgbus
import six
import logging
import sysactions

from sysactions import set_custom
from fiscalinterface import FiscalProcessor
from xml.etree import cElementTree as eTree
from threading import Thread, Condition
from msgbus import MBException, FM_STRING, FM_PARAM, TK_SYS_ACK, MBEasyContext
from syserrors import SE_NOTFOUND, SE_MB_CONNECT
from systools import sys_log_exception
from bustoken import (
    TK_SAT_PROCESS_REQUEST,
    TK_SAT_STATUS_REQUEST,
    TK_SAT_CANCEL_SALE,
    TK_SAT_CANCEL_ORDER
)
from old_helper import convert_from_localtime_to_utc
from pos_model import OrderParser
from satbuilder import CfeBuilder
from datetime import datetime
from common import FiscalParameterController
from pos_util import SaleLineUtil

logger = logging.getLogger("FiscalWrapper")


class SatRequestBuilder:
    def __init__(self, mbcontext, crt, cnpj_sw, sign_ac, cnpj_contribuinte, inscr_estadual, order_parser, fiscal_parameter_controller, sale_line_util, cfe_builder, is_mfe):
        self.crt = crt
        self.mbcontext = mbcontext
        self.cnpj_sw = cnpj_sw
        self.sign_ac = sign_ac
        self.cnpj_contribuinte = cnpj_contribuinte
        self.inscr_estadual = inscr_estadual
        self.order_parser = order_parser  # type: OrderParser
        self.fiscal_parameter_controller = fiscal_parameter_controller  # type: FiscalParameterController
        self.sale_line_util = sale_line_util  # type: SaleLineUtil
        self.cfe_builder = cfe_builder  # type: CfeBuilder
        self.is_mfe = is_mfe

    def get_sat_info(self):
        return self.cnpj_sw, self.sign_ac

    def build_request(self, order_xml):
        # type: (eTree) -> unicode
        order = self.order_parser.parse_order(order_xml)

        context = {}
        xml = self.cfe_builder.build_xml(order, context)
        return xml


class SatService:
    def __init__(self, service_name):
        self.service_name = service_name
        self.available = True


class SatServiceFinder:
    def __init__(self, mbcontext, max_sat_number, pos_sat_list, sleep_time):
        # type: (MBEasyContext, int, str, int) -> None
        self.mbcontext = mbcontext

        self.max_sat_number = max_sat_number
        self.pos_sat_list = pos_sat_list.split(';') if pos_sat_list else []
        self.sleep_time = sleep_time

        self.search_thread = None  # type: Optional[Thread]
        self.search_thread_condition = Condition()
        self.search_thread_running = False
        self.available_sats = {}  # type: Dict[str, SatService]
        self.available_sats_keys = []
        self.next_sat_index = 0
        self.available_sats_lock = Condition()

        self.service_name_initial = "SAT"

        self.search_sat_modules()
        self.start_search_thread()

    def find_and_lock_sat_service(self, order_id):
        # type: (str) -> str
        with self.available_sats_lock:
            if not self.available_sats:
                with self.search_thread_condition:
                    self.search_thread_condition.notify()
                raise Exception("Nenhum SAT encontrado")

            initial_sat_index = self.next_sat_index

            try:
                sat_service_name = self.available_sats_keys[self.next_sat_index]
            except IndexError:
                logger.exception("Erro ao acessar SAT com indice %d. Sats disponiveis: %d" % (self.next_sat_index, len(self.available_sats_keys)))
                self.next_sat_index = 0
                initial_sat_index = len(self.available_sats_keys)
                sat_service_name = self.available_sats_keys[self.next_sat_index]
            else:
                self.next_sat_index += 1
                if self.next_sat_index >= len(self.available_sats):
                    self.next_sat_index = 0

            while not self.available_sats[sat_service_name].available:
                sat_service_name = self.available_sats_keys[self.next_sat_index]
                self.next_sat_index += 1
                if self.next_sat_index >= len(self.available_sats):
                    self.next_sat_index = 0

                if self.next_sat_index == initial_sat_index:
                    # Todos os SATs disponiveis estao ocupados, vamos esperar alguem liberar algum
                    self.available_sats_lock.wait()

                    # Alguem liberou um sat, vamos tentar pega-lo
                    initial_sat_index = self.next_sat_index
                    sat_service_name = self.available_sats_keys[self.next_sat_index]

            self.available_sats[sat_service_name].available = False
            return sat_service_name

    def unlock_sat_service(self, sat_service_name):
        # type: (str) -> None
        with self.available_sats_lock:
            if sat_service_name in self.available_sats:
                self.available_sats[sat_service_name].available = True

                # Como liberamos um SAT, vamos notificar para caso alguem esteja esperando
                self.available_sats_lock.notify()

    def sat_unavailable(self, sat_service_name):
        # type: (str) -> None
        if len(self.available_sats) == 1:
            return
        with self.available_sats_lock:
            if sat_service_name in self.available_sats:
                del self.available_sats[sat_service_name]
                self.available_sats_keys.remove(sat_service_name)
                posid = int(sat_service_name[3:])
                set_custom(posid, 'SAT_STATUS', '2')

                # Removemos um SAT - Vamos arrumar os indices
                self._fix_sat_index()

    def search_sat_modules(self):
        logger.debug("Start Search SAT Function")
        sat_list_to_search = self.pos_sat_list if self.pos_sat_list else range(0, self.max_sat_number + 1)

        now_available = {}
        for i in sat_list_to_search:
            sat_service_name = "SAT%s" % str(i).zfill(2)

            try:
                ret = self.mbcontext.MB_EasySendMessage(sat_service_name, TK_SAT_STATUS_REQUEST, format=FM_PARAM, timeout=100 * 100000)
                if ret.data == "OK":
                    now_available[sat_service_name] = SatService(sat_service_name)
                    try:
                        set_custom(i, 'SAT_STATUS', '1') if i != 0 else None
                    except Exception:
                        pass
                else:
                    if ret.data != "No SAT attached":
                        logger.warning("{0} not available: {1}".format(sat_service_name, ret.data))
                    try:
                        set_custom(i, 'SAT_STATUS', '2') if i != 0 else None
                    except Exception:
                        pass
            except MBException as e:
                try:
                    set_custom(i, 'SAT_STATUS', '3') if i != 0 else None
                except Exception:
                    pass
                if e.errorcode not in [SE_NOTFOUND, SE_MB_CONNECT]:
                    logger.exception(e.message)

        with self.available_sats_lock:
            # Primeiro removemos os que estavam disponiveis e agora nao estao mais
            sats_to_remove = []
            for sat_service_name in self.available_sats:  # type: str
                if sat_service_name not in now_available:
                    sats_to_remove.append(sat_service_name)

            for sat_service_name in sats_to_remove:
                del self.available_sats[sat_service_name]
                self.available_sats_keys.remove(sat_service_name)

            # Agora adicionamos os que passaram a estar disponiveis
            sats_novos = False
            for sat_service_name in now_available:  # type: str
                if sat_service_name not in self.available_sats:
                    sat = now_available[sat_service_name]  # type: SatService

                    self.available_sats[sat.service_name] = sat
                    self.available_sats_keys.append(sat.service_name)

                    sats_novos = True

            # Removemos um SAT - Vamos arrumar os indices
            self._fix_sat_index()

            if sats_novos:
                # Como temos SATs novos disponiveis, vamos avisar caso algume thread esteja esperando
                self.available_sats_lock.notify()
        logger.debug("End Search SAT Function")

    def start_search_thread(self):
        self.search_thread_running = True
        self.search_thread = Thread(target=self._search_thread_loop)
        self.search_thread.daemon = True
        self.search_thread.start()

    def stop_search_thread(self):
        if self.search_thread_running:
            with self.search_thread_condition:
                self.search_thread_running = False
                self.search_thread_condition.notifyAll()

    def _search_thread_loop(self):
        while self.search_thread_running:
            with self.search_thread_condition:
                if not self.available_sats:
                    self.search_thread_condition.wait(15)
                else:
                    self.search_thread_condition.wait(self.sleep_time)
            self.search_sat_modules()

    def _fix_sat_index(self):
        # Atualizamos o indice, pois houve alguma remoção de SATs
        if self.next_sat_index >= len(self.available_sats):
            self.next_sat_index = len(self.available_sats) - 1

            # Não permirte que o índice fique negativo
            if self.next_sat_index < 0:
                self.next_sat_index = 0


class MfeServiceFinder(SatServiceFinder):
    def find_and_lock_sat_service(self, order_id):
        # type: (str) -> str
        with self.available_sats_lock:
            if not self.available_sats:
                raise Exception("Nenhum SAT encontrado")

            sat_service_name = self.available_sats_keys[int(order_id) % len(self.available_sats_keys)]

            while not self.available_sats[sat_service_name].available:
                self.available_sats_lock.wait()

            self.available_sats[sat_service_name].available = False
            return sat_service_name


class SatProcessor(FiscalProcessor):
    def __init__(self, mbcontext, sat_service_finder, sat_request_builder, fiscal_sent_dir):
        # type: (msgbus.MBEasyContext, SatServiceFinder, SatRequestBuilder, str) -> SatProcessor
        super(SatProcessor, self).__init__()

        self.mbcontext = mbcontext
        self.sat_service_finder = sat_service_finder
        self.sat_request_builder = sat_request_builder
        self.last_sat_used = {}
        self.fiscal_sent_dir = fiscal_sent_dir
        sysactions.mbcontext = self.mbcontext
        self.cancel_data = None

    def get_sat_info(self):
        return self.sat_request_builder.get_sat_info()

    def request_fiscal(
            self,
            pos_id,             # type: str
            order,              # type: eTree
            tenders,            # type: List[Tender]
            paf=False,          # type: Optional[bool]
            fiscal_id=None      # type: str
    ):
        # type: (...) -> eTree.Element

        request = self.sat_request_builder.build_request(order)

        sat_service_name = self.sat_service_finder.find_and_lock_sat_service(order.get("orderId"))

        self.last_sat_used[pos_id] = sat_service_name

        try:
            order_id = order.get("orderId")
            order_id = order_id.zfill(9)

            dir_enviadas = os.path.join(self.fiscal_sent_dir, "Enviados")
            if not os.path.exists(dir_enviadas):
                os.makedirs(dir_enviadas)
            req = '\0'.join([pos_id, order.attrib["orderId"], request.encode('utf-8'), fiscal_id])
            msg = self.mbcontext.MB_EasySendMessage(sat_service_name, TK_SAT_PROCESS_REQUEST, format=FM_STRING, data=req)

            self.sat_service_finder.unlock_sat_service(sat_service_name)

            if msg.token == TK_SYS_ACK:
                ret_xml = eTree.XML(msg.data)

                element = ret_xml.find("XmlRequest").text
                # Arrumamos o padding, que pode ter vindo errado e o base64 exige que o mesmo esteja com padding correto
                padded = element + "=" * ((4 - len(element) % 4) % 4)
                logger.debug("OrderId: %s; request size: %d" % (order_id, len(padded)))
                base64_xml = base64.b64decode(padded)
                logger.debug("OrderId: %s; base64_xml: %s" % (order_id, base64_xml))
                req_xml = eTree.XML(base64_xml)

                num_nota_cfe_el = req_xml.find("infCFe/ide/nCFe")
                num_nota_cfe = num_nota_cfe_el.text.zfill(9) if num_nota_cfe_el is not None else '000000000'
                data_emissao = req_xml.find("infCFe/ide/dEmi").text
                hora_emissao = req_xml.find("infCFe/ide/hEmi").text
                emissao_date = convert_from_localtime_to_utc(datetime.strptime(data_emissao + hora_emissao, "%Y%m%d%H%M%S"))
                dir_nota = os.path.join(dir_enviadas, data_emissao[0:4], data_emissao[4:6], data_emissao[6:8])
                if not os.path.exists(dir_nota):
                    os.makedirs(dir_nota)

                self.order_service.set_custom_property(
                    pos_id=pos_id,
                    key="FISCALIZATION_DATE",
                    value=emissao_date.strftime("%Y-%m-%dT%H:%M:%S"),
                    blk_notify=True,
                )

                self.sat_service_finder.unlock_sat_service(sat_service_name)

                cfe_code = ret_xml.find("FiscalData").find("satkey").text

                response_file_dir = os.path.join(dir_nota, "S{0}_{1}_{2}_response_pos{3}_CFe{4}.xml".format(sat_service_name[-2:], num_nota_cfe, order_id, str(pos_id).zfill(2), cfe_code))
                logger.debug("OrderId: %s; response_file_dir: %s" % (order_id, response_file_dir))
                with open(response_file_dir, "w+") as response_file:
                    response_file.write(base64_xml)
                return ret_xml
            else:
                dir_erro = os.path.join(self.fiscal_sent_dir, "Erros")
                if not os.path.exists(dir_erro):
                    os.makedirs(dir_erro)
                # Salvamos o arquivo com erro em arquivo, no formato request_sat<satname>_pos<posid>_order<orderid>_erro
                request_erro = open(os.path.join(dir_erro, "{0}_request_{1}_pos{2}_order{3}_erro.xml".format(order_id, sat_service_name, str(pos_id).zfill(2), order.get("orderId"))), "w+")
                request_erro.write(str(request))
                request_erro.close()
                raise Exception(msg.data)

        except Exception as e:
            sys_log_exception(e.message)
            self.sat_service_finder.sat_unavailable(sat_service_name)
            raise six.reraise(Exception, e, sys.exc_info()[2])

    def do_validation(self, get_days_to_expiration=None):
        return True, "OK"

    def get_fiscal_sent_dir(self):
        return self.fiscal_sent_dir

    def terminate(self):
        self.sat_service_finder.stop_search_thread()

    def cancel_sale(
        self,
        order_data,     # type: str
        order_id,       # type: str
    ):
        # type: (...) -> Dict

        self.cancel_data = None
        sat_service = "SAT" + order_data[0]
        cancel_data = self.build_cancel_xml(order_data[1])
        msg = self.mbcontext.MB_EasySendMessage(sat_service, TK_SAT_CANCEL_SALE, format=FM_PARAM, timeout=100 * 100000, data=str(cancel_data))
        if msg.token == TK_SYS_ACK:
            self.cancel_data = msg.data
            ret_xml = eTree.XML(msg.data)
            element = ret_xml.find("XmlResponse").text
            message = ret_xml.find("Message").text
            # Se veio XML Response, arrumamos o padding, que pode ter vindo errado e o base64 exige que o mesmo esteja com padding correto e salvamos em arquivo
            if element:
                padded = element + "=" * ((4 - len(element) % 4) % 4)
                base64_xml = base64.b64decode(padded)
                cfe = cancel_data[1]
                num_nota_cfe = cfe[34:40].zfill(9)
                idx_start = base64_xml.find("<dEmi>") + 6
                idx_final = base64_xml.find("</dEmi>")
                data_emissao = base64_xml[idx_start:idx_final]
                dir_arquivo = os.path.join(data_emissao[0:4], data_emissao[4:6], data_emissao[6:8])
                pos_id = sat_service[-2:]
                file_name = "S{}_{}_{}_cancelamento_{}.xml".format(pos_id, num_nota_cfe, order_id.zfill(9), cfe)
                file_path = os.path.join(self.fiscal_sent_dir, "Enviados", dir_arquivo, file_name)
                with open(file_path, "w+") as cfe_proc_file:
                    cfe_proc_file.write(base64_xml)

                return {'status': True, 'message': message, 'xml_response': padded}

            return {'status': True, 'message': message, 'xml_response': None}
        else:
            return {'status': False, 'message': msg.data, 'xml_response': None}

    def build_cancel_xml(self, order_request):
        sale_request_xml = eTree.XML(base64.b64decode(order_request + "==="))

        chcanc = sale_request_xml.find("infCFe").attrib.get("Id")
        cnpj = self.sat_request_builder.cnpj_sw
        signac = self.sat_request_builder.sign_ac
        posid = sale_request_xml.find("infCFe").find("ide").find("numeroCaixa").text

        cancel_xml = "<CFeCanc>" \
                     "<infCFe chCanc=\"{chcanc}\">" \
                     "<ide>" \
                     "<CNPJ>{cnpj}</CNPJ>" \
                     "<signAC>{signac}</signAC>" \
                     "<numeroCaixa>{posid}</numeroCaixa>" \
                     "</ide>" \
                     "<emit/>" \
                     "<dest/>" \
                     "<total/>" \
                     "</infCFe>" \
                     "</CFeCanc>" \
            .format(chcanc=chcanc, cnpj=cnpj, signac=signac, posid=posid)
        return cancel_xml, chcanc



from logging import Logger

from helper import json_serialize
from interactor.route.order.change_items_quantity.dto import ChangeItemsQuantityRequest
from model.config import Config
from model.error_type import PosValidationErrorsType
from model.exception import ValidationError
from model.repository import PosRepositoryModel
from validation_util import \
    check_if_business_day_is_opened, \
    check_if_operator_is_logged, \
    check_if_order_is_in_progress, \
    validate_api_key


class ChangeItemsQuantityInteractor(object):
    def __init__(self, logger, config, pos_repository):
        # type: (Logger, Config, PosRepositoryModel) -> None

        self.logger = logger
        self.config = config
        self.pos_repository = pos_repository

    def execute(self, request):
        # type: (ChangeItemsQuantityRequest) -> None

        pos_id = validate_api_key(request, self.config.api_key_mapping)
        self.logger.info("Changing items quantity. PosId: {}; Items: {}"
                         .format(pos_id, json_serialize(request.items)))

        if not request.items:
            raise ValidationError(PosValidationErrorsType.SaleItemsInvalid)

        model = self.pos_repository.get_model(pos_id)

        check_if_business_day_is_opened(self.pos_repository, model)
        check_if_operator_is_logged(self.pos_repository, model)
        check_if_order_is_in_progress(self.pos_repository, model)

        pos_ot = self.pos_repository.get_pos_ot(model)

        self.pos_repository.change_line_quantity(model, pos_ot, request.items)



from abc import ABCMeta, abstractmethod

from _Store import Store  # noqa


class StoreRetriever(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_store(self):
        # type: () -> Store
        raise NotImplementedError()



class SaleLine(object):
    def __init__(self, seat_number, order_id, line_number, item_id, level, part_code):
        self.seat_number = seat_number
        self.order_id = order_id
        self.line_number = line_number
        self.item_id = item_id
        self.level = level
        self.part_code = part_code

    @classmethod
    def load_info(cls, seat_number, sale_line):
        cls.seat_number = seat_number
        cls.order_id = sale_line["orderId"] if "orderId" in sale_line else None
        cls.line_number = sale_line["lineNumber"] if "lineNumber" in sale_line else None
        cls.item_id = sale_line["itemId"] if "itemId" in sale_line else None
        cls.level = sale_line["level"] if "level" in sale_line else None
        cls.part_code = sale_line["partCode"] if "partCode" in sale_line else None
        return cls



# defusedxml
#
# Copyright (c) 2013 by Christian Heimes <christian@python.org>
# Licensed to PSF under a Contributor Agreement.
# See https://www.python.org/psf/license for licensing details.
"""Defused xml.dom.minidom
"""
from __future__ import print_function, absolute_import

from xml.dom.minidom import _do_pulldom_parse
from . import expatbuilder as _expatbuilder
from . import pulldom as _pulldom

__origin__ = "xml.dom.minidom"


def parse(
    file, parser=None, bufsize=None, forbid_dtd=False, forbid_entities=True, forbid_external=True
):
    """Parse a file into a DOM by filename or file object."""
    if parser is None and not bufsize:
        return _expatbuilder.parse(
            file,
            forbid_dtd=forbid_dtd,
            forbid_entities=forbid_entities,
            forbid_external=forbid_external,
        )
    else:
        return _do_pulldom_parse(
            _pulldom.parse,
            (file,),
            {
                "parser": parser,
                "bufsize": bufsize,
                "forbid_dtd": forbid_dtd,
                "forbid_entities": forbid_entities,
                "forbid_external": forbid_external,
            },
        )


def parseString(
    string, parser=None, forbid_dtd=False, forbid_entities=True, forbid_external=True
):
    """Parse a file into a DOM from a string."""
    if parser is None:
        return _expatbuilder.parseString(
            string,
            forbid_dtd=forbid_dtd,
            forbid_entities=forbid_entities,
            forbid_external=forbid_external,
        )
    else:
        return _do_pulldom_parse(
            _pulldom.parseString,
            (string,),
            {
                "parser": parser,
                "forbid_dtd": forbid_dtd,
                "forbid_entities": forbid_entities,
                "forbid_external": forbid_external,
            },
        )



# -*- coding: utf-8 -*-

from datetime import datetime
from xml.etree import cElementTree as eTree

import sysactions
from msgbus import MBEasyContext
from old_helper import BaseRepository
from persistence import Connection
from typing import List, Optional

from ..model import Order


class OrderRepository(BaseRepository):
    TypeRealDate = "RealDate"
    TypeBusinessPeriod = "BusinessPeriod"
    TypeSessionId = "SessionId"

    def __init__(self, mbcontext, pos_list):
        # type: (MBEasyContext, List[int]) -> None
        super(OrderRepository, self).__init__(mbcontext)
        self.pos_list = pos_list

    def get_discount_orders_by_real_date(self, initial_date, end_date, operator_id, report_pos):
        # type: (datetime, datetime, unicode, unicode) -> List[Order]
        return self._get_discount_orders(self.TypeRealDate, initial_date, end_date, operator_id, None, report_pos)

    def get_discount_orders_by_business_period(self, initial_date, end_date, operator_id, report_pos):
        # type: (datetime, datetime, unicode, unicode) -> List[Order]
        return self._get_discount_orders(self.TypeBusinessPeriod, initial_date, end_date, operator_id, None, report_pos)

    def get_discount_orders_by_session_id(self, session_id):
        # type: (unicode) -> List[Order]
        return self._get_discount_orders(self.TypeSessionId, None, None, None, session_id, None)

    def _get_discount_orders(self, report_type, initial_date, end_date, operator_id, session_id, report_pos):
        # type: (unicode, Optional[datetime], Optional[datetime], Optional[unicode], Optional[unicode], Optional[unicode]) -> List[Order]
        def inner_func(conn):
            # type: (Connection) -> List[Order]

            date_format = "%Y%m%d"
            if operator_id is not None:
                current_session_id = "user={},".format(operator_id)
            else:
                current_session_id = session_id if session_id is not None and session_id != 'None' else ''

            if report_pos == self.TypeSessionId:
                query = self._DiscountOrdersBySessionIdQuery.format(session_id)
            else:
                if report_type == self.TypeBusinessPeriod:
                    query = self._DiscountOrdersByBusinessPeriodQuery
                else:
                    query = self._DiscountOrdersByRealDateQuery
                query = query.format(initial_date.strftime(date_format), end_date.strftime(date_format), current_session_id)

            order_with_payments = [(x.get_entry(0),
                                    x.get_entry(1),
                                    x.get_entry(2),
                                    x.get_entry(3),
                                    x.get_entry(4),
                                    x.get_entry(5),
                                    x.get_entry(6),
                                    x.get_entry(7)) for x in conn.select(query)]

            all_orders = []  # type: List[Order]
            for order_params in order_with_payments:
                is_delivery = bool(int(order_params[7]))
                order_id = int(order_params[0])
                discount_amount = float(order_params[1])
                total_gross = float(order_params[2])
                applied_date = datetime.strptime(order_params[3], "%Y-%m-%dT%H:%M:%S") if not is_delivery else None
                operator = self.get_operator_name(order_params[4] or '') if not is_delivery else None
                authorizer = self.get_operator_name(order_params[5] or '') if not is_delivery else None
                delivery = is_delivery

                order = Order(order_id, discount_amount, total_gross, applied_date, operator, authorizer, delivery)
                all_orders.append(order)

            return all_orders

        report_pos_list = self.pos_list if report_pos is None else (report_pos, )
        return self.execute_parallel_in_all_databases_returning_flat_list(inner_func, report_pos_list)

    @staticmethod
    def get_operator_name(operator_id):
        return eTree.XML(sysactions.get_user_information(int(operator_id))).find(".//user").get("LongName").upper()

    _DiscountOrdersByRealDateQuery = \
"""
    select 
        o.orderid, 
        o.discountamount,
        o.totalgross,
        ocp1.value as AppliedDate,
        ocp2.value as Operator,
        ocp3.value as Authorizer,
        ocp4.value as Method,
        ocp5.value IS NOT NULL as Delivery
    from orders o
    left join ordercustomproperties ocp1 on ocp1.orderid = o.orderid and ocp1.key = 'DISCOUNT_APPLY_DATETIME'
    left join ordercustomproperties ocp2 on ocp2.orderid = o.orderid and ocp2.key = 'DISCOUNT_APPLY_OPERATOR'
    left join ordercustomproperties ocp3 on ocp3.orderid = o.orderid and ocp3.key = 'DISCOUNT_APPLY_AUTHORIZER'
    left join ordercustomproperties ocp4 on ocp4.orderid = o.orderid and ocp4.key = 'DISCOUNT_APPLY_METHOD'
    left join ordercustomproperties ocp5 on ocp5.orderid = o.orderid and ocp5.key = 'REMOTE_ORDER_ID'
    inner join (
        select orderid, strftime('%Y%m%d', value, 'localtime') OrderDate
        from ordercustomproperties
        where key = 'FISCALIZATION_DATE' 
            and strftime('%Y%m%d', value, 'localtime') >= '{0}' 
            and strftime('%Y%m%d', value, 'localtime') <= '{1}'
    ) od on o.orderid = od.orderid
    where cast(o.discountamount as float) <> 0 
    and o.stateid = 5
    and o.sessionid like '%{2}%'
"""

    _DiscountOrdersByBusinessPeriodQuery = \
"""
    select 
        o.orderid, 
        o.discountamount,
        o.totalgross,
        ocp1.value as AppliedDate,
        ocp2.value as Operator,
        ocp3.value as Authorizer,
        ocp4.value as Method,
        ocp5.value IS NOT NULL as Delivery
    from orders o
    left join ordercustomproperties ocp1 on ocp1.orderid = o.orderid and ocp1.key = 'DISCOUNT_APPLY_DATETIME'
    left join ordercustomproperties ocp2 on ocp2.orderid = o.orderid and ocp2.key = 'DISCOUNT_APPLY_OPERATOR'
    left join ordercustomproperties ocp3 on ocp3.orderid = o.orderid and ocp3.key = 'DISCOUNT_APPLY_AUTHORIZER'
    left join ordercustomproperties ocp4 on ocp4.orderid = o.orderid and ocp4.key = 'DISCOUNT_APPLY_METHOD'
    left join ordercustomproperties ocp5 on ocp5.orderid = o.orderid and ocp5.key = 'REMOTE_ORDER_ID'
	where o.businessperiod >= '{0}' and o.businessperiod <= '{1}' 
        and o.stateid = 5
        and cast(o.discountamount as float) <> 0 
        and o.sessionid like '%{2}%'
"""

    _DiscountOrdersBySessionIdQuery = \
"""
    select 
        o.orderid, 
        o.discountamount,
        o.totalgross,
        ocp1.value as AppliedDate,
        ocp2.value as Operator,
        ocp3.value as Authorizer,
        ocp4.value as Method,
        ocp5.value IS NOT NULL as Delivery
    from orders o
    left join ordercustomproperties ocp1 on ocp1.orderid = o.orderid and ocp1.key = 'DISCOUNT_APPLY_DATETIME'
    left join ordercustomproperties ocp2 on ocp2.orderid = o.orderid and ocp2.key = 'DISCOUNT_APPLY_OPERATOR'
    left join ordercustomproperties ocp3 on ocp3.orderid = o.orderid and ocp3.key = 'DISCOUNT_APPLY_AUTHORIZER'
    left join ordercustomproperties ocp4 on ocp4.orderid = o.orderid and ocp4.key = 'DISCOUNT_APPLY_METHOD'
    left join ordercustomproperties ocp5 on ocp5.orderid = o.orderid and ocp5.key = 'REMOTE_ORDER_ID'
	where o.sessionid = '{0}' 
        and o.stateid = 5 
        and cast(o.discountamount as float) <> 0 
"""



from typing import List


class ListenedEvents(object):
    DISCOUNT_APPLIED = "DISCOUNT_APPLIED"
    ORDER_CANCELED = "ORDER_CANCELED"
    COMMON_CLEAN_DATABASES = "COMMON_CLEAN_DATABASES"
    POS_ALERT_NOTIFICATION_ACK = "PosAlertNotificationAck"
    BOH_CHANGE_NOTIFICATION_CONFIG = "BohChangeNotification"
    BOH_ORDER_PICTURE = "BOH_ORDER_PICTURE"


def all_events():
    # type: (...) -> List[str]

    listened_events = [
        ListenedEvents.DISCOUNT_APPLIED,
        ListenedEvents.ORDER_CANCELED,
        ListenedEvents.COMMON_CLEAN_DATABASES,
        ListenedEvents.POS_ALERT_NOTIFICATION_ACK,
        ListenedEvents.BOH_CHANGE_NOTIFICATION_CONFIG,
        ListenedEvents.BOH_ORDER_PICTURE,
    ]

    return listened_events



from typing import Dict  # noqa
from commons.report import Generator
from .dto import CashReportDto, PeriodCashReportFilterDto, CashReportBodyDto, CashReportBodyLineDto
from domain import OperatorNameRetriever, StoreRetriever, OrderRepository, Order, \
    TransferRepository, Transfer, SessionRepository, Clock  # noqa
from commons.dto import DefaultHeaderDto


class PeriodCashReportGenerator(Generator):
    def __init__(self,
                 filter,
                 clock,
                 operator_name_retriever,
                 store_retriever,
                 order_repository,
                 transfer_repository,
                 session_repository):
        # type: (PeriodCashReportFilterDto, Clock, OperatorNameRetriever, StoreRetriever, OrderRepository, TransferRepository, SessionRepository) -> None  # noqa
        self.filter = filter
        self.clock = clock
        self.operator_name_retriever = operator_name_retriever
        self.store_retriever = store_retriever
        self.order_repository = order_repository
        self.transfer_repository = transfer_repository
        self.session_repository = session_repository

    def generate_data(self):
        # type: () -> CashReportDto
        header = self.generate_header()
        body = self.generate_body()
        return CashReportDto(header, body)

    def generate_body(self):
        body = CashReportBodyDto()
        body.operator_count = self.calculate_operator_count()
        self.fill_order_fields(body)
        self.fill_transfer_fields(body)
        body.value_in_drawer = self.calculate_value_in_drawer(body)
        return body

    def generate_header(self):
        report_type = self.filter.report_type
        store_code = self.store_retriever.get_store().code
        operator_name = None
        if self.filter.operator_id is not None:
            operator_name = self.operator_name_retriever.get_operator_name(self.filter.operator_id)
        generated_time = self.clock.get_current_datetime()
        return DefaultHeaderDto(report_type,
                                self.filter.initial_date,
                                self.filter.end_date,
                                self.filter.pos,
                                self.filter.operator_id,
                                operator_name,
                                store_code,
                                generated_time)

    def calculate_operator_count(self):
        return self.session_repository.get_unique_user_session_count(
            self.filter.initial_date,
            self.filter.end_date)

    def fill_order_fields(self, body):
        tender_dict = {}  # type: Dict[str, CashReportBodyLineDto]
        orders = self.order_repository.get_order(self.filter.initial_date,
                                                 self.filter.end_date,
                                                 self.filter.pos,
                                                 self.filter.operator_id)
        for order in orders:
            self.add_order_to_field(body.total_sales, order)

            if order.state == Order.OrderSate.voided:
                self.add_order_to_field(body.voids, order)

            if order.state == Order.OrderSate.paid:
                self.add_order_to_field(body.net_sales, order)

            for tender in order.tenders:
                self.add_tender_to_dict(tender, tender_dict)
        body.tender_breakdown = list(tender_dict.values())

    @staticmethod
    def add_order_to_field(field, order, detail=None):
        # type: (CashReportBodyLineDto, Order, str) -> None  # noqa
        field.quantity += 1
        field.value += order.total
        field.detail = detail

    @staticmethod
    def add_tender_to_dict(tender, tender_dict):
        if tender.type not in tender_dict:
            tender_dict[tender.type] = CashReportBodyLineDto(0, 0.0, tender.type)
        tender_dict[tender.type].quantity += 1
        tender_dict[tender.type].value += tender.value

    def fill_transfer_fields(self, body):
        transfers = self.transfer_repository.get_transfers(self.filter.initial_date,
                                                           self.filter.end_date)
        for transfer in transfers:
            if transfer.type == Transfer.Type.initial_float:
                self.add_transfer_to_field(body.initial_float, transfer)

            if transfer.type == Transfer.Type.cash_in:
                self.add_transfer_to_field(body.cash_ins, transfer)

            if transfer.type in (Transfer.Type.cash_out, Transfer.Type.last_cash_out):
                self.add_transfer_to_field(body.cash_outs, transfer)

    @staticmethod
    def add_transfer_to_field(field, transfer):
        # type: (CashReportBodyLineDto, Transfer) -> None
        field.quantity += 1
        field.value += transfer.value

    @staticmethod
    def calculate_value_in_drawer(body):
        value = body.initial_float.value + body.cash_ins.value
        for tender in body.tender_breakdown:
            if tender.detail == "MONEY":
                value += tender.value
        value -= body.cash_outs.value

        return value



# encoding: utf-8
from typing import Dict


class BenefitsBodyDto(object):
    def __init__(self, orders, type_report):
        # type: (Dict, str) -> ()

        self.items = orders
        self.type_report = type_report



class SubscriptionInfo(object):
    TYPE_REENTRANT = 1
    TYPE_NON_REENTRANT = 2

    def __init__(self, subject, subscription_type):
        # type: (str, int) -> None
        self.subject = subject
        self.subscription_type = subscription_type



from msgbus import TK_PRN_PRINT

from messagebus import (
    DataType,
    Message,
    MessageBus,
)


class PrinterService(object):
    def __init__(
        self,
        message_bus,    # type: MessageBus
        timeout,        # type: int
    ):
        # type: (...) -> None

        self.message_bus = message_bus
        self.timeout = timeout

    def send_to_printer(
        self,
        printer_name,       # type: str
        text,               # type: str
    ):
        #  type: (...) -> Message

        message = Message(
            token=TK_PRN_PRINT,
            data_type=DataType.string,
            data=text,
            timeout=self.timeout,
        )

        return self.message_bus.send_message(
            component=printer_name,
            message=message,
        )




import logging
import logging.config
import logging.handlers
import os
import sys
import json

import persistence
import unicodedata
from enum import Enum
from customencodings import custom_encode
from typing import (
    Union,
    Optional,
)


def config_logger(
    loader_path,                    # type: str
    log_name='component',           # type: Optional[str]
    configure_root=False,           # type: Optional[bool]
    root_level=logging.WARN,        # type: Optional[int]
    max_files=5,                    # type: Optional[int]
    max_size=5242880,               # type: Optional[int]
    add_error_handler=False,        # type: Optional[bool]
):
    # type: (...) -> None

    real_path = os.path.realpath(loader_path)
    if os.path.isfile(real_path) and real_path.endswith(".ini"):
        logging.config.fileConfig(real_path, disable_existing_loggers=False)
        return

    if os.path.isfile(real_path) and real_path.endswith(".json"):
        with open(real_path, "rb") as f:
            logging.config.dictConfig(json.loads(f.read().decode("utf-8")))
        return

    if os.path.isdir(real_path):
        dir_path = real_path
    else:
        dir_path = os.path.dirname(real_path)

    filename = os.path.join(dir_path, log_name + '.log')
    filename_error = os.path.join(dir_path, log_name + '.error.log')

    log_string = "%(asctime)-6s: %(name)s - %(levelname)s - %(thread)d - %(threadName)s - %(message)s"
    formatter = logging.Formatter(log_string)

    comp_logger = logging.getLogger(log_name)
    comp_logger.setLevel(logging.DEBUG)
    comp_logger.propagate = False

    rotating_file_logger = logging.handlers.RotatingFileHandler(filename, 'a', max_size, max_files)
    rotating_file_logger.setLevel(logging.DEBUG)
    rotating_file_logger.setFormatter(formatter)
    comp_logger.addHandler(rotating_file_logger)

    if add_error_handler:
        rotating_file_logger_error = logging.handlers.RotatingFileHandler(filename_error, 'a', max_size, max_files)
        rotating_file_logger_error.setLevel(logging.ERROR)
        rotating_file_logger_error.setFormatter(formatter)
        comp_logger.addHandler(rotating_file_logger_error)

    if configure_root:
        root_logger = logging.getLogger()
        root_logger.addHandler(rotating_file_logger)
        root_logger.setLevel(root_level)


def import_pydevd(loader_path, port, suspend=False):
    # type: (str, int, bool) -> None
    if "DEBUG" in os.environ:
        dir_path = os.path.dirname(os.path.realpath(loader_path))
        filename = os.path.join(dir_path, "debug.txt")
        if os.path.isfile(filename):
            with open(filename) as debug_file:
                debug_path = debug_file.read()
            if os.path.exists(debug_path):
                if debug_path not in sys.path:
                    sys.path.append(debug_path)
                try:
                    import pydevd
                    pydevd.settrace('localhost', port=port, stdoutToServer=True, stderrToServer=True, suspend=suspend)
                except ImportError:
                    pydevd_using_debug_lib(
                        loader_path=loader_path,
                        port=port,
                        suspend=suspend,
                    )
            else:
                debug_path = os.path.abspath(debug_path)
                print(debug_path)
                print("path errado")


def pydevd_using_debug_lib(
    loader_path,        # type: str
    port,               # type: int
    suspend=False,      # type: Optional[bool]
):
    # type: (...) -> None

    try:
        current_dir = os.path.dirname(os.path.abspath(__file__))
        debug_lib_dir = os.path.join(current_dir, "..", "debug_lib")
        if os.path.exists(debug_lib_dir):
            sys.path.append(debug_lib_dir)
            from debughelper import import_pydevd as real_import_pydevd
            real_import_pydevd(loader_path, port, suspend)
    except ImportError:
        print("error configure debug, run install_lib_debug.sh")
        return


def get_valid_pricelists(mbcontext):
    price_lists = []

    conn = None
    try:
        conn = persistence.Driver().open(mbcontext)
        cursor = conn.select("SELECT PriceListID FROM Pricelist WHERE CURRENT_DATE BETWEEN EnabledFrom AND EnabledThru")
        for row in cursor:
            pl_id = str(row.get_entry(0) or '')
            price_lists.append(pl_id)
    finally:
        if conn:
            conn.close()
    return price_lists


def unicode_2_ascii(data):
    import unicodedata

    try:
        data = data.decode('latin_1', 'ignore')
    except Exception: # noqa
        pass
    finally:
        data = unicode(data)
        data = unicodedata.normalize('NFKD', data)
        data = data.encode('ascii', 'ignore')

    return data


def remove_accents(text):
    # type: (Union[str, unicode]) -> str

    if not text:
        return text

    try:
        text = unicode(text, 'utf-8')
    except (TypeError, NameError):
        pass

    text = unicodedata.normalize('NFD', text)
    text = text.encode('ascii', 'ignore')
    return text


def encode_cp860(
    text    # type: str
):
    # type: (...) -> str

    try:
        unicode_text = unicode(text, "utf-8")
    except TypeError:
        unicode_text = text

    try:
        response_text = unicode_text.encode("cp860")
    except TypeError:
        response_text = remove_accents(unicode_text)

    return response_text


def custom_encode_cp860(
    text,    # type: str
):
    # type: (...) -> str

    try:
        return custom_encode(
            text=text,
        )
    except Exception as ex:
        return remove_accents(
            text=text,
        )


def format_cpf_cnpj(value):
    formatted_value = value
    try:
        if len(value) == 11:
            value_to_format = value.zfill(11)
            formatted_value = '{}.{}.{}-{}'.format(value_to_format[:3],
                                                   value_to_format[3:6],
                                                   value_to_format[6:9],
                                                   value_to_format[9:])
        elif len(value) == 14:
            value_to_format = value.zfill(14)
            formatted_value = '{}.{}.{}/{}-{}'.format(value_to_format[:2],
                                                      value_to_format[2:5],
                                                      value_to_format[5:8],
                                                      value_to_format[8:12],
                                                      value_to_format[12:])
    except Exception as _:
        return value

    return formatted_value


class ExtendedEnum(Enum):

    @classmethod
    def list_values(cls):
        # type: () -> list
        return list(map(lambda enum: enum.value, cls))

    @classmethod
    def list_names(cls):
        # type: () -> list
        return list(map(lambda enum: enum.name, cls))

    @classmethod
    def list_enums_tuple(cls):
        # type: () -> list
        return list(map(lambda enum: (enum.name, enum.value), cls))

    @classmethod
    def get_name(cls, value):
        # type: (int) -> Union[str, None]
        if value not in cls.list_values():
            return None

        return next(str(enum.name) for enum in cls if enum.value == value)

    @classmethod
    def get_value(cls, name):
        # type: (str) -> Union[int, None]
        if name not in cls.list_names():
            return None

        return next(int(enum.value) for enum in cls if enum.name == name)


class MwOrderStatus(ExtendedEnum):
    IN_PROGRESS = 1
    STORED = 2
    TOTALED = 3
    VOIDED = 4
    PAID = 5
    RECALLED = 6



# -*- coding: utf-8 -*-
import sysactions

from actions.login.signout_user import sign_out_user
from actions.util.list_users import get_opened_users


@sysactions.action
def close_operators(pos_id):
    model = sysactions.get_model(pos_id)
    sysactions.check_current_order(pos_id, model=model, need_order=False)

    pos_to_close_user, user, operator = get_opened_users(pos_id=pos_id)

    if user is None:
        return False

    return sign_out_user(
        pos_id=pos_id,
        user_id=user.get("id"),
        pos_to_close_user=pos_to_close_user,
    )



import collections
from datetime import datetime
from threading import Lock
from uuid import uuid4

from .. import Waiter

from ._PendingWaiter import PendingWaiter
from ._ResolvedWaiter import ResolvedWaiter


class Subscriber:
    def __init__(self):
        self.queue = collections.deque()
        self.last_seen = datetime.now()
        self.sync_id = str(uuid4())
        self.waiter_lock = Lock()
        self.pending_waiter = None  # type: PendingWaiter

    def new_event(self, event):
        with self.waiter_lock:
            if self.pending_waiter is not None:
                self.pending_waiter.new_event(event)
                self.pending_waiter = None
                self.last_seen = datetime.now()
            else:
                self.queue.append(event)

    def get_waiter(self):
        # type: () -> Waiter
        self.sync_id = str(uuid4())

        events = []
        with self.waiter_lock:
            if len(self.queue) > 0:
                try:
                    while True:
                        events.append(self.queue.popleft())
                except IndexError:
                    pass
            else:
                self.pending_waiter = PendingWaiter(self.sync_id)

        if len(events) > 0:
            return ResolvedWaiter(self.sync_id, events)
        else:
            return self.pending_waiter



from helper import ExtendedEnum
from messagebus import TokenCreator, TokenPriority

MESSAGE_GROUP = "111"


class ListenedTokens(ExtendedEnum):
    TK_LOYALTY_GET_LOYALTY_ID = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "1")
    TK_LOYALTY_GET_AND_LOCK_VOUCHER = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "2")
    TK_LOYALTY_BURN_VOUCHER = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "3")
    TK_LOYALTY_UNLOCK_VOUCHER = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "4")
    TK_LOYALTY_IS_AVAILABLE = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "FF")



# encoding: utf-8
from old_helper import BaseRepository

from ..model import OrderTender


class OrderRepository(BaseRepository):

    def __init__(self, mbcontext, pos_list):
        super(OrderRepository, self).__init__(mbcontext)
        self.pos_list = pos_list

    def get_paid_orders_by_tender_type(self, session_id):
        def inner_func(conn):
            query = self._PaidOrdersTotalAmountSumQuery.format(session_id)
            orders = []
            for x in conn.select(query):
                orders.append(OrderTender(x.get_entry("TenderId"),
                                          x.get_entry("TypeDescr"),
                                          float(x.get_entry("TenderAmount") or 0.0)))
            return orders

        return self.execute_parallel_in_all_databases_returning_flat_list(inner_func, self.pos_list)

    _PaidOrdersTotalAmountSumQuery = """SELECT ot.TenderId, st.TypeDescr, SUM(COALESCE(TenderAmount, 0) - COALESCE(ChangeAmount, 0)) AS TenderAmount
                                        FROM Orders o
                                        INNER JOIN OrderTender ot ON o.OrderId = ot.OrderId
                                        LEFT JOIN SaleType st ON o.SaleType = st.TypeId
                                        WHERE o.StateId = 5 AND o.SessionId = '{}'
                                        GROUP BY st.TypeId, ot.TenderId"""



from typing import (
    Dict,
    List,
    Optional,
)

import sysactions
from posot import OrderTaker


from utils import (
    get_void_reason_id,
    get_canceled_by,
    do_confirm_void,
    get_fixed_void_reason,
)
from custom_sysactions import user_authorized_to_execute_action
from manager.reports import justification_report
from utilconfigurations import must_require_void_reason
from utilfunctions import (
    is_stored_order,
    is_order_in_progress,
)


@sysactions.action
def do_void_lines(
    pos_id,             # type: str
    line_numbers=None,  # type: Optional[str]
    auth_data=None,     # type: Optional[str]
):
    # type: (...) -> bool

    model = sysactions.get_model(posid=pos_id)
    stored_order = is_stored_order(model=model)
    in_progress_order = is_order_in_progress(model=model)

    if stored_order:
        do_void_lines_stored_order(
            pos_id,
            line_numbers=line_numbers,
            auth_data=auth_data,
        )
    elif in_progress_order:
        do_void_lines_in_progress_order(
            pos_id,
            line_numbers=line_numbers,
            auth_data=auth_data,
        )

    return True


@user_authorized_to_execute_action
def do_void_lines_in_progress_order(
    pos_id,         # type: str
    line_numbers,   # type: str
    auth_data,      # type: str
):
    # type: (...) -> None

    _void_lines(
        pos_id=pos_id,
        line_numbers=line_numbers,
        auth_data=auth_data,
    )


@user_authorized_to_execute_action
def do_void_lines_stored_order(
    pos_id,         # type: str
    line_numbers,   # type: str
    auth_data,      # type: str
):
    # type: (...) -> None

    _void_lines(
        pos_id=pos_id,
        line_numbers=line_numbers,
        auth_data=auth_data,
        stored_order=True,
    )


def _void_lines(
    pos_id,              # type: str
    line_numbers,        # type: str
    auth_data,           # type: str
    stored_order=False,  # type: Optional[bool]
):
    # type: (...) -> bool

    if not line_numbers:
        return False

    model = sysactions.get_model(pos_id)
    sysactions.check_current_order(pos_id, model=model)

    if not sysactions.can_void_line(model, line_numbers):
        sysactions.show_messagebox(pos_id, "$NEED_TO_HAVE_ITEM_TO_VOID")
        return False

    auth_user = auth_data if auth_data is not None else None
    canceled_by = get_canceled_by(
        pos_id,
        model,
        auth_user
    )

    void_type = "VOID_LINE_ON_STORED_ORDER" if stored_order else "VOID_LINE"
    if must_require_void_reason(void_type):
        void_reason = get_void_reason_id(
            pos_id=pos_id,
            void_type=void_type,
            authentication_user=canceled_by,
            is_void_line=True,
        )
        if void_reason is None:
            return True
    else:
        do_confirm_void(
            pos_id=pos_id,
            message="$CONFIRM_VOID_LINE",
        )
        void_reason = get_fixed_void_reason(
            reason="USER_CANCELLATION",
            canceled_by=canceled_by,
            void_type=void_type,
        )

    pos_ot = sysactions.get_posot(model)
    order = model.find("CurrentOrder/Order")
    order_id = order.get("orderId")
    sale_lines = order.findall("SaleLine")

    do_void_lines_justification_report(
        pos_id=pos_id,
        void_reason=void_reason,
        canceled_by=canceled_by,
        order_id=order_id,
        stored_order=stored_order,
    )

    pos_ot.voidLine(int(pos_id), line_numbers)
    update_lines_properties(
        pos_ot=pos_ot,
        order_id=order_id,
        line_numbers=line_numbers,
        void_reason=void_reason,
        canceled_by=canceled_by,
        sale_lines=sale_lines,
    )

    return True


def do_void_lines_justification_report(
    pos_id,         # type: str
    void_reason,    # type: Dict
    canceled_by,    # type: str
    order_id,       # type: str
    stored_order,   # type: bool
):
    # type: (...) -> None

    justification_type = "VOID_LINE_ON_STORED_ORDER" if stored_order else "VOID_LINE"
    justification_report(
        pos_id=pos_id,
        justification_type=justification_type,
        justification=void_reason['VOID_REASON_DESCR'],
        authorized_by=canceled_by,
        order_id=order_id,
    )


def update_lines_properties(
    pos_ot,         # type: OrderTaker
    order_id,       # type: str
    line_numbers,   # type: str
    void_reason,    # type: Dict
    canceled_by,    # type: str
    sale_lines,     # type: List
):
    # type: (...) -> None

    for line_number in line_numbers.split('|'):
        custom_properties = {}
        custom_properties.update(void_reason or "")
        if void_reason is None:
            custom_properties.update({"VOID_AUTH_USER": canceled_by})

        current_line_sale_lines = [x for x in sale_lines if x.get("lineNumber") == line_number]
        for current_line in current_line_sale_lines:
            item_type = current_line.get("itemType")
            if item_type == "OPTION":
                continue

            voided_qty = float(current_line.get("qty", 0))
            if item_type != "PRODUCT":
                voided_qty = voided_qty - float(current_line.get("defaultQty", 0))
            custom_properties.update({"VOID_QTY": str(voided_qty)})

            pos_ot.setOrderCustomProperties(
                custom_properties,
                orderid=order_id,
                linenumber=line_number,
                itemid=current_line.get("itemId"),
                level=current_line.get("level"),
                partcode=current_line.get("partCode"),
            )



from typing import Dict


class CustomizationStyle(object):
    def __init__(
            self,
            style   # type: Dict
    ):
        # type: (...) -> CustomizationStyle
        self.style = style



# -*- coding: utf-8 -*-

import logging
from typing import Dict


class DeliveryDiscounts(object):
    def __init__(
        self,
        logger,     # type: logging
    ):
        # type: (...) -> None

        self.logger = logger

    def get_discount_amount_by_type(
        self,
        vouchers,           # type: Dict
        discount_type,      # type: str
    ):
        # type: (...) -> float

        if not vouchers:
            return 0

        discount_vouchers = vouchers.get(discount_type)
        if not discount_vouchers:
            return 0

        discount_amount = 0

        for voucher_discount_type in discount_vouchers:
            if voucher_discount_type == 'delivery_fee':
                continue

            try:
                discount_amount += float(discount_vouchers.get(voucher_discount_type))
            except Exception as _:
                self.logger.info("Unhandled discount {}".format(voucher_discount_type))

        return discount_amount

    def get_delivery_fee_discounts(
        self,
        vouchers,   # type: Dict
    ):
        # type: (...) -> float

        if not vouchers:
            return 0

        discount_amount = 0
        for _, value in vouchers.items():
            delivery_fee = value.get('delivery_fee')
            if delivery_fee:
                try:
                    discount_amount += float(value.get('delivery_fee', 0))
                except Exception as _:
                    self.logger.info("Unhandled discount delivery_fee")

        return discount_amount



from datetime import datetime

from application.model import BenefitAppliers, OperationDescription
from typing import Optional


class BenefitControllerDto(object):

    def __init__(self):
        # type: () -> None

        self.benefit_id = None  # type: Optional[str]
        self.originator_pos_id = None  # type: Optional[int]
        self.operation_description = None  # type: Optional[OperationDescription]
        self.benefit_applier = None  # type: Optional[BenefitAppliers]
        self.retry_quantity = None  # type: Optional[int]
        self.last_retry_date_utc = None  # type: Optional[datetime]



"""
``python-future``: pure Python implementation of Python 3 round().
"""
 
from future.utils import PYPY, PY26, bind_method

# Use the decimal module for simplicity of implementation (and
# hopefully correctness).
from decimal import Decimal, ROUND_HALF_EVEN


def newround(number, ndigits=None):
    """
    See Python 3 documentation: uses Banker's Rounding.
 
    Delegates to the __round__ method if for some reason this exists.
 
    If not, rounds a number to a given precision in decimal digits (default
    0 digits). This returns an int when called with one argument,
    otherwise the same type as the number. ndigits may be negative.
 
    See the test_round method in future/tests/test_builtins.py for
    examples.
    """
    return_int = False
    if ndigits is None:
        return_int = True
        ndigits = 0
    if hasattr(number, '__round__'):
        return number.__round__(ndigits)
    
    if ndigits < 0:
        raise NotImplementedError('negative ndigits not supported yet')
    exponent = Decimal('10') ** (-ndigits)

    if PYPY:
        # Work around issue #24: round() breaks on PyPy with NumPy's types
        if 'numpy' in repr(type(number)):
            number = float(number)

    if not PY26:
        d = Decimal.from_float(number).quantize(exponent,
                                            rounding=ROUND_HALF_EVEN)
    else:
        d = from_float_26(number).quantize(exponent, rounding=ROUND_HALF_EVEN)

    if return_int:
        return int(d)
    else:
        return float(d)
 

### From Python 2.7's decimal.py. Only needed to support Py2.6:

def from_float_26(f):
    """Converts a float to a decimal number, exactly.

    Note that Decimal.from_float(0.1) is not the same as Decimal('0.1').
    Since 0.1 is not exactly representable in binary floating point, the
    value is stored as the nearest representable value which is
    0x1.999999999999ap-4.  The exact equivalent of the value in decimal
    is 0.1000000000000000055511151231257827021181583404541015625.

    >>> Decimal.from_float(0.1)
    Decimal('0.1000000000000000055511151231257827021181583404541015625')
    >>> Decimal.from_float(float('nan'))
    Decimal('NaN')
    >>> Decimal.from_float(float('inf'))
    Decimal('Infinity')
    >>> Decimal.from_float(-float('inf'))
    Decimal('-Infinity')
    >>> Decimal.from_float(-0.0)
    Decimal('-0')

    """
    import math as _math
    from decimal import _dec_from_triple    # only available on Py2.6 and Py2.7 (not 3.3)

    if isinstance(f, (int, long)):        # handle integer inputs
        return Decimal(f)
    if _math.isinf(f) or _math.isnan(f):  # raises TypeError if not a float
        return Decimal(repr(f))
    if _math.copysign(1.0, f) == 1.0:
        sign = 0
    else:
        sign = 1
    n, d = abs(f).as_integer_ratio()
    # int.bit_length() method doesn't exist on Py2.6:
    def bit_length(d):
        if d != 0:
            return len(bin(abs(d))) - 2
        else:
            return 0
    k = bit_length(d) - 1
    result = _dec_from_triple(sign, str(n*5**k), -k)
    return result


__all__ = ['newround']



from mw_helper import ensure_iterable

from ._OrderCondition import OrderCondition
from production.model import ProductionOrder


class OrderStateCondition(OrderCondition):
    def __init__(self, order_states):
        order_states = ensure_iterable(order_states)

        self.valid_order_states = {}
        for order_state in order_states:
            self.valid_order_states[order_state] = order_state

    def evaluate(self, order):
        # type: (ProductionOrder) -> bool

        return order.state in self.valid_order_states
